import os
import sys
import warnings

# _ROOT = os.path.abspath(os.path.dirname(__file__))
#
#
# def get_data_path(path):
#     return os.path.join(_ROOT, 'data', path)

# Setup warnings to simpler one line warning

def warning_on_one_line(message, category, filename, lineno, file=None, line=None):
    
    filename = os.path.join('pyrealm', os.path.basename(filename))
    return '%s:%s: %s: %s\n' % (filename, lineno, category.__name__, message)

warnings.formatwarning = warning_on_one_line

# # And provide a decorator to catch warnings in doctests
# # https://stackoverflow.com/questions/2418570/

# def stderr_to_stdout(func):
#     def wrapper(*args):
#         stderr_bak = sys.stderr
#         sys.stderr = sys.stdout
#         try:
#             return func(*args)
#         finally:
#             sys.stderr = stderr_bak
#     return wrapper


import numpy as np
from scipy.optimize import curve_fit
import warnings
import matplotlib.pyplot as plt
import json


# https://stackoverflow.com/questions/51662473/using-gaussian-mixture-for-1d-array-in-python-sklearn

# Differences between this and Maxime's work
# 1) Gaussian curves estimated simultaneously. Maxime split the data frame
#    and estimated separate curves for each partition.
# 2) Selection using AICc to identify the best representation


def gauss_single(x: np.ndarray, ampl: float, mu: float, sigma:float) -> np.ndarray:
    """Calculates the density of a Gaussian distribution for values in
    ``x`` given the amplitude, mean and variance.

    Args:
        x: An array of values to calculate t
        ampl: The amplitude
        mu: The mean
        sigma: The variance

    Returns:
        An ndarray with the same shape as ``x`` containing density values
    """

    return ampl * np.exp(-(x - mu) ** 2 / (2. * sigma ** 2))


def gauss_multiple(x, *pars):
    """Calculates the density for values in ``x`` of a set of superimposed
     Gaussian distributions. The parameters of those distributions are passed
     in ``pars`` as a sequence of parameters giving the amplitude, mean and
     variance in turn of each Gaussian. This allows the function to be used
     easily with curve_fit.

    Args:
        x: An array of values to calculate t
        *pars: Gaussian parameters

    Returns:
        An ndarray with the same shape as ``x`` containing density values
    """
    if len(x) <= len(pars):
        raise ValueError('Overfitting: more parameters than data')
    try:
        pars = np.array(pars)
        pars = pars.reshape(-1, 3)
    except ValueError:
        raise ValueError('gaussN requires extra parameters in multiples of 3')

    curves = np.array([gauss_single(x, *pr) for pr in pars])

    return curves.sum(axis=0)


def calc_log_lik(obs: np.ndarray, pred: np.ndarray) -> float:
    """Calculates the log-likelihood of the predicted data given the
    observations

    Args:
        obs: An array of observed values
        pred: n array of predicted values of the same shape as ``obs``

    Returns:
        The log likelihood estimate
    """

    n_obs = len(obs)
    res = obs - pred

    return 0.5 * (- n_obs * (np.log(2 * np.pi) + 1 - np.log(n_obs) +
                             np.log(sum(res ** 2))))


def calc_aic(obs: np.ndarray, pred: np.ndarray, params: np.ndarray,
             k: int = 2, aicc: bool = True) -> float:
    """

    Args:
        obs: An array of observed values
        pred: An array of predicted values of the same shape as ``obs``
        params: An array of the estimated parameters used to predict ``pred``
        k: Penalty value.
        aicc: Apply the small sample correction to AIC.

    Returns:
        The AIC or AICc value
    """

    ll = calc_log_lik(obs, pred)
    df = len(params) + 1
    no = len(obs)

    if aicc:
        return (-2 * ll) + (k * df) * (1 + ((df + 1)/(no - df - 1)))
    else:
        return (-2 * ll) + (k * df)


class GaussSelect:

    def __init__(self, x, y):

        self.x = x
        self.y = y

        # Define attributes used in fit() method
        self.nmax = None
        self.use_bounds = None
        self.use_aicc = None
        self.aic = None
        self.fitted = None
        self.fit_complete = False

    def fit(self, n_max=3, use_aicc=True, use_bounds=True):

        self.use_aicc = use_aicc
        self.nmax = n_max
        self.aic = np.empty(n_max + 1)
        self.aic[:] = np.nan
        self.fitted = [None] * (n_max + 1)
        self.use_bounds = use_bounds

        # Null model - no gaussians
        self.fitted[0] = []
        self.aic[0] = calc_aic(self.y, np.ones_like(self.y) * self.y.mean(), self.fitted[0],
                               aicc=self.use_aicc)

        # Loop over gaussian fits
        for ng in np.arange(1, n_max + 1):

            # Initial estimates - provided to the function as a 1D array
            # of contiguous sets of amplitude, mean and sigma
            init = np.ones(3 * ng)

            # Bound the amplitude and sigma to be positive, but not the mean
            if self.use_bounds:
                lbound = (0, -np.inf, 0) * ng
                ubound = (np.inf, np.inf, np.inf) * ng
                bounds = (lbound, ubound)
            else:
                bounds = (-np.inf, np.inf)

            try:
                # Fit the curve and save the estimated parameters
                self.fitted[ng], _ = curve_fit(gauss_multiple, self.x, self.y, p0=init,
                                               bounds=bounds)

                # Get the predictions for this number of gaussians
                yp = gauss_multiple(self.x, self.fitted[ng])

                # Estimate AIC
                self.aic[ng] = calc_aic(self.y, yp, self.fitted[ng],
                                        aicc=self.use_aicc)

            except RuntimeError as err:
                warnings.warn(f'Could not fit multiple gaussian for n = {ng}: {err}')

        self.fit_complete = True

    def plot(self, res=101):

        if self.fit_complete:
            # Interpolate closely spaced points across x
            xplot = np.linspace(np.min(self.x), np.max(self.x), res)

            # Get predictions for null model and each gaussian fit
            # interpolating NaN for models that did not fit.
            yplots = [np.ones_like(xplot) * self.y.mean()]

            for cf in self.fitted[1:]:

                if cf is not None:
                    yplots.append(gauss_multiple(xplot, cf))
                else:
                    no_fit = np.empty_like(xplot)
                    no_fit[:] = np.nan
                    yplots.append(no_fit)

            # Plot worse models in grey
            best_fit = np.nanargmin(self.aic)
            worse_models = np.delete(np.arange(len(self.fitted)), best_fit)

            for idx in worse_models:

                plt.plot(xplot, yplots[idx], 'grey')

            plt.plot(xplot, yplots[best_fit], 'red')

        plt.scatter(self.x, self.y)

        plt.show()




#
# mort = '/Users/dorme/Research/REALM/Files_from_Maxime/Nouveau dossier/sukumar_mort.json'
# recr = '/Users/dorme/Research/REALM/Files_from_Maxime/Nouveau dossier/sukumar_recr.json'
#
# mort_data = json.load(open(mort))
# recr_data = json.load(open(recr))
#
# species = ["LAGL", "TERT", "HELI", "ANOL", "TECG", "TECG",
#            "RAND", "SYZC", "PHYE", "KYDC", "STET", "GRET"]
#
# size_class_bounds = np.array([0, 0.9993767, 1.5991664, 2.5628244, 4.1307247,
#                               6.7151617, 11.0502286, 18.5878071, 32.429648,
#                               59.8845962, 117.383105, 235.2833097, 480])
#
# size_class_widths = np.diff(size_class_bounds, n=1)
# size_class_mids = size_class_bounds[:-1] + size_class_widths / 2
# ln_size_class_mids = np.log(size_class_mids)
#
#
# for sp in species:
#
#     sp_mort_data = mort_data[sp]['survivors']
#     sp_recr_data = recr_data[sp]['recruits.or.dead']
#
#     first = [rw['1988-1992'] for rw in sp_mort_data]
#
#     rates = first[:-1] / size_class_widths
#
#     g = GaussSelect(ln_size_class_mids, rates)
#     g.plot()
#     g.fit(use_bounds=False)
#     g.plot()
#
#
# mort_data['GRET']['survivors']
#
# ## Components
#
# # T Model
# default_settings =
# {'a_hd': 116.0,
#  'ca_ratio': 390.43,
#  'h_max': 25.33,
#  'rho_s': 200.0,
#  'lai': 1.8,
#  'sla': 14.0,
#  'tau_f': 4.0,
#  'tau_r': 1.04,
#  'par_ext': 0.5,
#  'yld': 0.17,
#  'zeta': 0.17,
#  'resp_r': 0.913,
#  'resp_s': 0.044,
#  'resp_f': 0.1}


# pylint: disable=C0103
from typing import Optional, Union
import numpy as np
from warnings import warn
from pyrealm.utilities import summarize_attrs
from pyrealm.param_classes import PModelParams
from pyrealm.bounds_checker import bounds_checker
from pyrealm.utilities import check_input_shapes

# TODO - Note that the typing currently does not enforce the dtype of ndarrays
#        but it looks like the upcoming np.typing module might do this.

def calc_density_h2o(tc: Union[float, np.ndarray],
                     patm: Union[float, np.ndarray],
                     pmodel_params: PModelParams = PModelParams(),
                     safe: bool = True) -> Union[float, np.ndarray]:
    """Calculates the **density of water** as a function of temperature and
    atmospheric pressure, using the Tumlirz Equation and coefficients calculated
    by :cite:`Fisher:1975tm`.

    Parameters:

        tc: air temperature, °C
        patm: atmospheric pressure, Pa
        pmodel_params: An instance of :class:`~pyrealm.param_classes.PModelParams`.
        safe: Prevents the function from estimating density below -30°C, where the
            function behaves poorly

    Other Parameters:

        lambda_: polynomial coefficients of Tumlirz equation (`pmodel_params.fisher_dial_lambda`).
        Po: polynomial coefficients of Tumlirz equation (`pmodel_params.fisher_dial_Po`).
        Vinf: polynomial coefficients of Tumlirz equation (`pmodel_params.fisher_dial_Vinf`).

    Returns:

        Water density as a float in (g cm^-3)

    Examples:

        >>> round(calc_density_h2o(20, 101325), 4)
        998.2056
    """

    # DESIGN NOTE:
    # It doesn't make sense to use this function for tc < 0, but in particular
    # the calculation shows wild numeric instability between -44 and -46 that
    # leads to numerous downstream issues - see the extreme values documentation.
    if safe and np.nanmin(tc) < -30:
        raise RuntimeError('Water density calculations below about -30°C are '
                           'unstable. See argument safe to calc_density_h2o')
        
    # Check input shapes, shape not used
    _ = check_input_shapes(tc, patm)

    # Get powers of tc, including tc^0 = 1 for constant terms
    tc_pow = np.power.outer(tc, np.arange(0, 10))

    # Calculate lambda, (bar cm^3)/g:
    lambda_val = np.sum(np.array(pmodel_params.fisher_dial_lambda) * tc_pow[..., :5], axis=-1)

    # Calculate po, bar
    po_val = np.sum(np.array(pmodel_params.fisher_dial_Po) * tc_pow[..., :5], axis=-1)

    # Calculate vinf, cm^3/g
    vinf_val = np.sum(np.array(pmodel_params.fisher_dial_Vinf) * tc_pow, axis=-1)

    # Convert pressure to bars (1 bar <- 100000 Pa)
    pbar = 1e-5 * patm

    # Calculate the specific volume (cm^3 g^-1):
    spec_vol = vinf_val + lambda_val / (po_val + pbar)

    # Convert to density (g cm^-3) -> 1000 g/kg; 1000000 cm^3/m^3 -> kg/m^3:
    rho = 1e3 / spec_vol

    # CDLO: Method of Chen et al (1997) - I tested this to compare to the TerrA-P
    # code base but I don't think we need it. Preserving the code in case it is
    # needed in the future.
    #
    #     # Calculate density at 1 atm (kg/m^3):
    #     chen_po = np.array([0.99983952, 6.788260e-5 , -9.08659e-6 , 1.022130e-7 ,
    #                         -1.35439e-9 , 1.471150e-11, -1.11663e-13, 5.044070e-16,
    #                         -1.00659e-18])
    #     po = np.sum(np.array(chen_po) * tc_pow[..., :9], axis=-1)
    #
    #     # Calculate bulk modulus at 1 atm (bar):
    #     chen_ko = np.array([19652.17, 148.1830, -2.29995, 0.01281,
    #                         -4.91564e-5, 1.035530e-7])
    #     ko = np.sum(np.array(chen_ko) * tc_pow[..., :6], axis=-1)
    #
    #     # Calculate temperature dependent coefficients:
    #     chen_ca = np.array([3.26138, 5.223e-4, 1.324e-4, -7.655e-7, 8.584e-10])
    #     ca = np.sum(np.array(chen_ca) * tc_pow[..., :5], axis=-1)
    #
    #     chen_cb = np.array([7.2061e-5, -5.8948e-6, 8.69900e-8, -1.0100e-9, 4.3220e-12])
    #     cb = np.sum(np.array(chen_cb) * tc_pow[..., :5], axis=-1)
    #
    #     # Convert atmospheric pressure to bar (1 bar = 100000 Pa)
    #     pbar = 1.0e-5 * patm
    #
    #     rho = (ko + ca * pbar + cb * pbar ** 2.0)
    #     rho /= (ko + ca * pbar + cb * pbar ** 2.0 - pbar)
    #     rho *= 1e3 * po

    return rho


def calc_ftemp_arrh(tk: Union[float, np.ndarray],
                    ha: float,
                    pmodel_params: PModelParams = PModelParams()
                    ) -> Union[float, np.ndarray]:
    r"""Calculates the temperature-scaling factor :math:`f` for enzyme kinetics
    following an Arrhenius response for a given temperature (``tk``, :math:`T`)
    and activation energy (`ha`, :math:`H_a`).

    Arrhenius kinetics are described as:

    .. math::

        x(T)= exp(c - H_a / (T R))

    The temperature-correction function :math:`f(T, H_a)` is:

      .. math::
        :nowrap:

        \[
            \begin{align*}
                f &= \frac{x(T)}{x(T_0)} \\
                  &= exp \left( \frac{ H_a (T - T_0)}{T_0 R T}\right)\text{, or equivalently}\\
                  &= exp \left( \frac{ H_a}{R} \cdot \left(\frac{1}{T_0} - \frac{1}{T}\right)\right)
            \end{align*}
        \]

    Parameters:

        tk: Temperature (in Kelvin)
        ha: Activation energy (in :math:`J \text{mol}^{-1}`)
        pmodel_params: An instance of :class:`~pyrealm.param_classes.PModelParams`.

    Other Parameters:

        To: a standard reference temperature (:math:`T_0`, `pmodel_params.k_To`)
        R: the universal gas constant (:math:`R`, `pmodel_params.k_R`)

    Returns:

        A float value for :math:`f`

    Examples:

        >>> # Relative rate change from 25 to 10 degrees Celsius (percent change)
        >>> round((1.0-calc_ftemp_arrh( 283.15, 100000)) * 100, 4)
        88.1991
    """

    # Note that the following forms are equivalent:
    # exp( ha * (tk - 298.15) / (298.15 * kR * tk) )
    # exp( ha * (tc - 25.0)/(298.15 * kR * (tc + 273.15)) )
    # exp( (ha/kR) * (1/298.15 - 1/tk) )

    tkref = pmodel_params.k_To + pmodel_params.k_CtoK

    return np.exp(ha * (tk - tkref) / (tkref * pmodel_params.k_R * tk))


def calc_ftemp_inst_rd(tc: Union[float, np.ndarray],
                       pmodel_params: PModelParams = PModelParams()
                       ) -> Union[float, np.ndarray]:
    """Calculates the **temperature-scaling factor for dark respiration**
    at a given temperature (``tc``, :math:`T` in °C), relative to the standard
    reference temperature :math:`T_o` (:cite:`Heskel:2016fg`).

    .. math::

            fr = exp( b (T_o - T) -  c ( T_o^2 - T^2 ))

    Parameters:

        tc: Temperature (degrees Celsius)

    Other parameters:

        To: standard reference temperature (:math:`T_o`, `pmodel_params.k_To`)
        b: empirically derived global mean coefficient (:math:`b`, Table 1, ::cite:`Heskel:2016fg`)
        c: empirically derived global mean coefficient (:math:`c`, Table 1, ::cite:`Heskel:2016fg`)


    Returns:

        A float value for :math:`fr`

    Examples:

        >>> # Relative percentage instantaneous change in Rd going from 10 to 25 degrees
        >>> val = (calc_ftemp_inst_rd(25) / calc_ftemp_inst_rd(10) - 1) * 100
        >>> round(val, 4)
        250.9593
    """

    return np.exp(pmodel_params.heskel_b * (tc - pmodel_params.k_To) -
                  pmodel_params.heskel_c * (tc ** 2 - pmodel_params.k_To ** 2))


def calc_ftemp_inst_vcmax(tc: Union[float, np.ndarray],
                          pmodel_params: PModelParams = PModelParams()
                          ) -> Union[float, np.ndarray]:
    r"""This function calculates the **temperature-scaling factor :math:`f` of
    the instantaneous temperature response of :math:`V_{cmax}`** given the
    temperature (:math:`T`) relative to the standard reference temperature
    (:math:`T_0`), following modified Arrhenius kinetics.

    .. math::

       V = f V_{ref}

    The value of :math:`f` is given by :cite:`Kattge:2007db` (Eqn 1) as:

    .. math::

        f = g(T, H_a) \cdot
                \frac{1 + exp( (T_0 \Delta S - H_d) / (T_0 R))}
                     {1 + exp( (T \Delta S - H_d) / (T R))}

    where :math:`g(T, H_a)` is a regular Arrhenius-type temperature response
    function (see :func:`calc_ftemp_arrh`). The term :math:`\Delta S` is the
    entropy factor, calculated as a linear function of :math:`T` in °C following
    :cite:`Kattge:2007db` (Table 3, Eqn 4):

    .. math::

        \Delta S = a + b T

    Parameters:

        tc:  temperature, or in general the temperature relevant for
            photosynthesis (°C)
        pmodel_params: An instance of :class:`~pyrealm.param_classes.PModelParams`.

    Other parameters:

        Ha: activation energy (:math:`H_a`, `pmodel_params.kattge_knorr_Ha`)
        Hd: deactivation energy (:math:`H_d`, `pmodel_params.kattge_knorr_Hd`)
        To: standard reference temperature expressed in Kelvin (`T_0`, `pmodel_params.k_To`)
        R: the universal gas constant (:math:`R`,`pmodel_params.k_R`)
        a: intercept of the entropy factor(:math:`a`, `pmodel_params.kattge_knorr_a_ent`)
        b: slope of the entropy factor (:math:`b`, `pmodel_params.kattge_knorr_b_ent`)

    Returns: A float value for :math:`f`

    Examples:

        >>> # Relative change in Vcmax going (instantaneously, i.e. not
        >>> # not acclimatedly) from 10 to 25 degrees (percent change):
        >>> val = ((calc_ftemp_inst_vcmax(25)/calc_ftemp_inst_vcmax(10)-1) * 100)
        >>> round(val, 4)
        283.1775

    """

    # Convert temperatures to Kelvin
    tkref = pmodel_params.k_To + pmodel_params.k_CtoK
    tk = tc + pmodel_params.k_CtoK

    # Calculate entropy following Kattge & Knorr (2007): slope and intercept
    # are defined using temperature in °C, not K!!! 'tcgrowth' corresponds
    # to 'tmean' in Nicks, 'tc25' is 'to' in Nick's
    dent = pmodel_params.kattge_knorr_a_ent + pmodel_params.kattge_knorr_b_ent * tc
    fva = calc_ftemp_arrh(tk, pmodel_params.kattge_knorr_Ha)
    fvb = ((1 + np.exp((tkref * dent - pmodel_params.kattge_knorr_Hd) /
                       (pmodel_params.k_R * tkref))) /
           (1 + np.exp((tk * dent - pmodel_params.kattge_knorr_Hd) /
                       (pmodel_params.k_R * tk))))

    return fva * fvb


def calc_ftemp_kphio(tc: Union[float, np.ndarray],
                     c4: bool = False,
                     pmodel_params: PModelParams = PModelParams()
                     ) -> Union[float, np.ndarray]:
    r"""Calculates the **temperature dependence of the quantum yield
    efficiency**, as a quadratic function of temperature (:math:`T`). The values
    of the coefficients depend on whether C3 or C4 photosynthesis is being
    modelled

    .. math::

        \phi(T) = a + b T - c T^2

    The factor :math:`\phi(T)` is to be multiplied with leaf absorptance and the
    fraction of absorbed light that reaches photosystem II. In the P-model these
    additional factors are lumped into a single apparent quantum yield
    efficiency parameter (argument `kphio` to the class :class:`~pyrealm.pmodel.PModel`).

    Parameters:

        tc: Temperature, relevant for photosynthesis (°C)
        c4: Boolean specifying whether fitted temperature response for C4 plants
            is used. Defaults to \code{FALSE}.
        pmodel_params: An instance of :class:`~pyrealm.param_classes.PModelParams`.

    Other parameters:

        C3: the parameters (:math:`a,b,c`, `pmodel_params.kphio_C3`) are taken from the
            temperature dependence of the maximum quantum yield of photosystem
            II in light-adapted tobacco leaves determined by :cite:`Bernacchi:2003dc`.
        C4: the parameters (:math:`a,b,c`, `pmodel_params.kphio_C4`) are taken from unpublished
            work.

    Returns:

        A float value for :math:`\phi(T)`

    Examples:

        >>> # Relative change in the quantum yield efficiency between 5 and 25
        >>> # degrees celsius (percent change):
        >>> val = (calc_ftemp_kphio(25.0) / calc_ftemp_kphio(5.0) - 1) * 100
        >>> round(val, 5)
        52.03969
        >>> # Relative change in the quantum yield efficiency between 5 and 25
        >>> # degrees celsius (percent change) for a C4 plant:
        >>> val = (calc_ftemp_kphio(25.0, c4=True) /
        ...        calc_ftemp_kphio(5.0, c4=True) - 1) * 100
        >>> round(val, 5)
        432.25806

    """

    if c4:
        coef = pmodel_params.kphio_C4
    else:
        coef = pmodel_params.kphio_C3

    ftemp = coef[0] + coef[1] * tc + coef[2] * tc ** 2
    ftemp = np.clip(ftemp, 0.0, None)
    
    return ftemp


def calc_gammastar(tc: Union[float, np.ndarray],
                   patm: Union[float, np.ndarray],
                   pmodel_params: PModelParams = PModelParams()
                   ) -> Union[float, np.ndarray]:
    r"""Calculates the photorespiratory **CO2 compensation point** in absence of
    dark respiration (:math:`\Gamma^{*}`, ::cite:`Farquhar:1980ft`) as:

    .. math::

        \Gamma^{*} = \Gamma^{*}_{0} \cdot \frac{p}{p_0} \cdot f(T, H_a)

    where :math:`f(T, H_a)` modifies the activation energy to the the local
    temperature following an Arrhenius-type temperature response function
    implemented in :func:`calc_ftemp_arrh`.

    Parameters:

        tc: Temperature relevant for photosynthesis (:math:`T`, °C)
        patm: Atmospheric pressure (:math:`p`, Pascals)
        pmodel_params: An instance of :class:`~pyrealm.param_classes.PModelParams`.

    Other Parameters:

        To: the standard reference temperature (:math:`T_0` )
        Po: the standard pressure (:math:`p_0` )
        gs_0: the reference value of :math:`\Gamma^{*} at standard temperature
            (:math:`T_0`) and pressure (:math:`P_0`)  (:math:`\Gamma^{*}_{0}`,
            ::cite:`Bernacchi:2001kg`, `pmodel_params.bernacchi_gs25_0`)
        ha: the activation energy (:math:`\Delta H_a`, ::cite:`Bernacchi:2001kg`,
            `pmodel_params.bernacchi_dha`)

    Returns:

        A float value for :math:`\Gamma^{*}` (in Pa)

    Examples:

        >>> # CO2 compensation point at 20 degrees Celsius and standard
        >>> # atmosphere (in Pa) >>> round(calc_gammastar(20, 101325), 5)
        3.33925
    """

    # check inputs, return shape not used
    _ = check_input_shapes(tc, patm)

    return (pmodel_params.bernacchi_gs25_0 * patm / pmodel_params.k_Po *
            calc_ftemp_arrh((tc + pmodel_params.k_CtoK), ha=pmodel_params.bernacchi_dha))


def calc_ns_star(tc: Union[float, np.ndarray],
                 patm: Union[float, np.ndarray],
                 pmodel_params: PModelParams = PModelParams()
                 ) -> Union[float, np.ndarray]:

    r"""Calculates the relative viscosity of water (:math:`\eta^*`), given the
    standard temperature and pressure, using :func:`~pyrealm.pmodel.calc_viscosity_h20`
    (:math:`v(t,p)`) as:

    .. math::

        \eta^* = \frac{v(t,p)}{v(t_0,p_0)}

    Parameters:

        tc: Temperature, relevant for photosynthesis (:math:`T`, °C)
        patm: Atmospheric pressure (:math:`p`, Pa)
        pmodel_params: An instance of :class:`~pyrealm.param_classes.PModelParams`.

    Other parameters:

        To: standard temperature (:math:`t0`, `pmodel_params.k_To`)
        Po: standard pressure (:math:`p_0`, `pmodel_params.k_Po`)

    Returns:

        A numeric value for :math:`K` (in Pa)

    Examples:

        >>> # Relative viscosity at 20 degrees Celsius and standard
        >>> # atmosphere (in Pa):
        >>> round(calc_ns_star(20, 101325), 5)
        1.12536
    """

    visc_env = calc_viscosity_h2o(tc, patm)
    visc_std = calc_viscosity_h2o(pmodel_params.k_To, pmodel_params.k_Po)

    return visc_env / visc_std


def calc_kmm(tc: Union[float, np.ndarray],
             patm: Union[float, np.ndarray],
             pmodel_params: PModelParams = PModelParams()
             ) -> Union[float, np.ndarray]:
    r"""Calculates the **Michaelis Menten coefficient of Rubisco-limited
    assimilation** (:math:`K`, ::cite:`Farquhar:1980ft`) as a function of
    temperature (:math:`T`) and atmospheric pressure (:math:`p`) as:

      .. math:: K = K_c ( 1 + p_{\ce{O2}} / K_o),

    where, :math:`p_{\ce{O2}} = 0.209476 \cdot p` is the partial pressure of
    oxygen. :math:`f(T, H_a)` is an Arrhenius-type temperature response of
    activation energies (:func:`calc_ftemp_arrh`) used to correct
    Michalis constants at standard temperature for both :math:`\ce{CO2}` and
    :math:`\ce{O2}` to the local temperature (Table 1, ::cite:`Bernacchi:2001kg`):

      .. math::
        :nowrap:

        \[
            \begin{align*}
                K_c &= K_{c25} \cdot f(T, H_{kc})\\
                K_o &= K_{o25} \cdot f(T, H_{ko})
            \end{align*}
        \]

    .. TODO - why this height? Inconsistent with calc_gammastar which uses P_0
              for the same conversion for a value in the same table.

    Parameters:

        tc: Temperature, relevant for photosynthesis (:math:`T`, °C)
        patm: Atmospheric pressure (:math:`p`, Pa)
        pmodel_params: An instance of :class:`~pyrealm.param_classes.PModelParams`.

    Other parameters:

        hac: activation energy for :math:`\ce{CO2}` (:math:`H_{kc}`, `pmodel_params.bernacchi_dhac`)
        hao:  activation energy for :math:`\ce{O2}` (:math:`\Delta H_{ko}`, `pmodel_params.bernacchi_dhao`)
        kc25: Michelis constant for :math:`\ce{CO2}` at standard temperature
            (:math:`K_{c25}`, `pmodel_params.bernacchi_kc25`)
        ko25: Michelis constant for :math:`\ce{O2}` at standard temperature
            (:math:`K_{o25}`, `pmodel_params.bernacchi_ko25`)

    Returns:

        A numeric value for :math:`K` (in Pa)

    Examples:

        >>> # Michaelis-Menten coefficient at 20 degrees Celsius and standard
        >>> # atmosphere (in Pa):
        >>> round(calc_kmm(20, 101325), 5)
        46.09928

    """

    # Check inputs, return shape not used
    _ = check_input_shapes(tc, patm)

    # conversion to Kelvin
    tk = tc + pmodel_params.k_CtoK

    kc = pmodel_params.bernacchi_kc25 * calc_ftemp_arrh(tk, ha=pmodel_params.bernacchi_dhac)
    ko = pmodel_params.bernacchi_ko25 * calc_ftemp_arrh(tk, ha=pmodel_params.bernacchi_dhao)

    # O2 partial pressure
    po = pmodel_params.k_co * 1e-6 * patm

    return kc * (1.0 + po/ko)


def calc_soilmstress(soilm: Union[float, np.ndarray],
                     meanalpha: Union[float, np.ndarray] = 1.0,
                     pmodel_params: PModelParams = PModelParams()
                     ) -> Union[float, np.ndarray]:
    r"""Calculates an **empirical soil moisture stress factor**  (:math:`\beta`,
    ::cite:`Stocker:2020dh`) as a function of relative soil moisture
    (:math:`m_s`, fraction of field capacity) and average aridity, quantified by
    the local annual mean ratio of actual over potential evapotranspiration
    (:math:`\bar{\alpha}`).

    The value of :math:`\beta` is defined relative to two soil moisture
    thresholds (:math:`\theta_0, \theta^{*}`) as:

      .. math::
        :nowrap:

        \[
            \beta =
                \begin{cases}
                    q(\theta_0 - \theta^{*})^2 + 1,  & \theta_0 < m_s <= \theta^{*} \\
                    1, &  \theta^{*} < m_s,
                \end{cases}
        \]

    where :math:`q` is an aridity sensitivity parameter setting the stress
    factor at :math:`\theta_0`:

    .. math:: q=(1 - (a + b \bar{\alpha}))/(\theta^{*} - \theta_{0})^2

    Default parameters are as described in :cite:`Stocker:2020dh`.

    Parameters:

        soilm: Relative soil moisture as a fraction of field capacity
            (unitless). Defaults to 1.0 (no soil moisture stress).
        meanalpha: Local annual mean ratio of actual over potential
            evapotranspiration, measure for average aridity. Defaults to 1.0.
        pmodel_params: An instance of :class:`~pyrealm.param_classes.PModelParams`.

    Other parameters:

        theta0: lower bound of soil moisture (:math:`\theta_0`, `pmodel_params.soilmstress_theta0`).
        thetastar: upper bound of soil moisture (:math:`\theta^{*}`, `pmodel_params.soilmstress_thetastar`).
        a: aridity parameter (:math:`a`, `pmodel_params.soilmstress_a`).
        b: aridity parameter (:math:`b`, `pmodel_params.soilmstress_b`).

    Returns:

        A numeric value for :math:`\beta`

    Examples:

        >>> # Relative reduction (%) in GPP due to soil moisture stress at
        >>> # relative soil water content ('soilm') of 0.2:
        >>> round((calc_soilmstress(0.2) - 1) * 100, 5)
        -14.0
    """

    # Check inputs, return shape not used
    _ = check_input_shapes(soilm, meanalpha)

    # Calculate outstress
    y0 = (pmodel_params.soilmstress_a + pmodel_params.soilmstress_b * meanalpha)
    beta = (1.0 - y0) / (pmodel_params.soilmstress_theta0 - pmodel_params.soilmstress_thetastar) ** 2
    outstress = 1.0 - beta * (soilm - pmodel_params.soilmstress_thetastar) ** 2

    # Filter wrt to thetastar
    outstress = np.where(soilm <= pmodel_params.soilmstress_thetastar, outstress, 1.0)

    # Clip
    outstress = np.clip(outstress, 0.0, 1.0)

    return outstress


def calc_viscosity_h2o(tc: Union[float, np.ndarray],
                       patm: Union[float, np.ndarray],
                       pmodel_params: PModelParams = PModelParams()
                       ) -> Union[float, np.ndarray]:
    r"""Calculates the **viscosity of water** (:math:`\eta`) as a function of
    temperature and atmospheric pressure (::cite:`Huber:2009fy`).

    Parameters:

        tc: air temperature (°C)
        patm: atmospheric pressure (Pa)
        pmodel_params: An instance of :class:`~pyrealm.param_classes.PModelParams`.

    Returns:

        A float giving the viscosity of water (mu, Pa s)

    Examples:

        >>> # Density of water at 20 degrees C and standard atmospheric pressure:
        >>> round(calc_viscosity_h2o(20, 101325), 7)
        0.0010016
    """

    # Check inputs, return shape not used
    _ = check_input_shapes(tc, patm)

    # Get the density of water, kg/m^3
    rho = calc_density_h2o(tc, patm, pmodel_params=pmodel_params)

    # Calculate dimensionless parameters:
    tbar = (tc + pmodel_params.k_CtoK) / pmodel_params.huber_tk_ast
    rbar = rho / pmodel_params.huber_rho_ast

    # Calculate mu0 (Eq. 11 & Table 2, Huber et al., 2009):
    tbar_pow = np.power.outer(tbar, np.arange(0, 4))
    mu0 = (1e2 * np.sqrt(tbar)) / np.sum(np.array(pmodel_params.huber_H_i) / tbar_pow, axis=-1)

    # Calculate mu1 (Eq. 12 & Table 3, Huber et al., 2009):
    h_array = np.array(pmodel_params.huber_H_ij)
    ctbar = (1.0 / tbar) - 1.0
    row_j, _ = np.indices(h_array.shape)
    mu1 = h_array * np.power.outer(rbar - 1.0, row_j)
    mu1 = np.power.outer(ctbar, np.arange(0, 6)) * np.sum(mu1, axis=(-2))
    mu1 = np.exp(rbar * mu1.sum(axis=-1))

    # Calculate mu_bar (Eq. 2, Huber et al., 2009), assumes mu2 = 1
    mu_bar = mu0 * mu1

    # Calculate mu (Eq. 1, Huber et al., 2009)
    return mu_bar * pmodel_params.huber_mu_ast  # Pa s


def calc_patm(elv: Union[float, np.ndarray],
              pmodel_params: PModelParams = PModelParams()
              ) -> Union[float, np.ndarray]:
    r"""Calculates **atmospheric pressure** as a function of elevation with reference to
    the standard atmosphere.  The elevation-dependence of atmospheric pressure
    is computed by assuming a linear decrease in temperature with elevation and
    a mean adiabatic lapse rate (Eqn 3, ::cite:`BerberanSantos:2009bk`):

    .. math::

        p(z) = p_0 ( 1 - L z / K_0) ^{ G M / (R L) },

    Parameters:

        elv: Elevation above sea-level (:math:`z`, metres above sea level.)
        pmodel_params: An instance of :class:`~pyrealm.param_classes.PModelParams`.

    Other parameters:

        G: gravity constant (:math:`g`, `pmodel_params.k_G`)
        Po: standard atmospheric pressure at sea level (:math:`p_0`, `pmodel_params.k_Po`)
        L: adiabatic temperature lapse rate (:math:`L}`, `pmodel_params.k_L`),
        M: molecular weight for dry air (:math:`M`, `pmodel_params.k_Ma`),
        R: universal gas constant (:math:`R`, `pmodel_params.k_R`)
        Ko: reference temperature in Kelvin (:math:`K_0`, `pmodel_params.k_To`).

    Returns:

        A numeric value for :math:`p` in Pascals.

    Examples:

        >>> # Standard atmospheric pressure, in Pa, corrected for 1000 m.a.s.l.
        >>> round(calc_patm(1000), 2)
        90241.54

    """

    # Convert elevation to pressure, Pa. This equation uses the base temperature
    # in Kelvins, while other functions use this constant in the PARAM units of
    # °C.

    kto = pmodel_params.k_To + pmodel_params.k_CtoK

    return (pmodel_params.k_Po * (1.0 - pmodel_params.k_L * elv / kto) **
            (pmodel_params.k_G * pmodel_params.k_Ma /
             (pmodel_params.k_R * pmodel_params.k_L)))


def calc_co2_to_ca(co2: Union[float, np.ndarray],
                   patm: Union[float, np.ndarray],
                   ) -> Union[float, np.ndarray]:
    r"""Converts ambient :math:`\ce{CO2}` (:math:`c_a`) in part per million to
    Pascals, accounting for atmospheric pressure.

    Parameters:
        co2 (float): atmospheric :math:`\ce{CO2}`, ppm
        patm (float): atmospheric pressure, Pa

    Returns:
        Ambient :math:`\ce{CO2}` in units of Pa

    Examples:
        >>> np.round(calc_co2_to_ca(413.03, 101325), 6)
        41.850265
    """


    return 1.0e-6 * co2 * patm  # Pa, atms. CO2


# Design notes on PModel (0.3.1 -> 0.4.0)
# The PModel until 0.3.1 was a single class taking tc etc. as inputs. However
# a common use case would be to look at how the PModel predictions change with
# different options. I (DO) did consider retaining the single class and having
# PModel __init__ create the environment and then have PModel.fit(), but
# having two classes seemed to better separate the physiological model (PModel
# class and attributes) from the environment the model is being fitted to and
# also creates _separate_ PModel objects.

# Design notes on PModel (0.4.0 -> 0.5.0)
# In separating PPFD and FAPAR into a second step, I had created the IabsScaled
# class to store variables that scale linearly with Iabs. That class held and
# exposed scaled and unscaled versions of several parameteres. For a start, that
# was a bad class name since the values could be unscaled, but also most of the
# unscaled versions are never really used. Really (hat tip, Keith Bloomfield),
# there are two meaningful _efficiency_ variables (LUE, IWUE) and then a set of
# productivity variables. The new structure reflects that, removing the
# un-needed unscaled variables and simplifying the structure.


class PModelEnvironment:

    """Create a PModelEnvironment instance using the input parameters.

    This class takes the four key environmental inputs to the P Model and
    calculates four photosynthetic variables for those environmental
    conditions:

    * the photorespiratory CO2 compensation point (:math:`\Gamma^{*}`,
      :func:`~pyrealm.pmodel.calc_gammastar`),
    * the relative viscosity of water (:math:`\eta^*`,
      :func:`~pyrealm.pmodel.calc_ns_star`),
    * the ambient partial pressure of :math:`\ce{CO2}` (:math:`c_a`,
      :func:`~pyrealm.pmodel.calc_c02_to_ca`) and
    * the Michaelis Menten coefficient of Rubisco-limited assimilation
      (:math:`K`, :func:`~pyrealm.pmodel.calc_kmm`).

    These variables can then be used to fit P models using different
    configurations. Note that the underlying parameters of the P model
    (:class:`~pyrealm.param_classes.PModelParams`) are set when creating
    an instance of this class.

    Parameters:

        tc: Temperature, relevant for photosynthesis (°C)
        vpd: Vapour pressure deficit (Pa)
        co2: Atmospheric CO2 concentration (ppm)
        patm: Atmospheric pressure (Pa).
        pmodel_params: An instance of :class:`~pyrealm.param_classes.PModelParams`.
    """

    def __init__(self,
                 tc: Union[float, np.ndarray],
                 vpd: Union[float, np.ndarray],
                 co2: Union[float, np.ndarray],
                 patm: Union[float, np.ndarray],
                 pmodel_params: PModelParams = PModelParams()):

        self.shape = check_input_shapes(tc, vpd, co2, patm)

        # Validate and store the forcing variables
        self.tc = bounds_checker(tc, -25, 80, '[]', 'tc', '°C')
        self.vpd = bounds_checker(vpd, 0, 10000, '[]', 'vpd', 'Pa')
        self.co2 = bounds_checker(co2, 0, 1000, '[]', 'co2', 'ppm')
        self.patm = bounds_checker(patm, 30000, 110000, '[]', 'tc', '°C')

        # Guard against calc_density issues
        if np.nanmin(self.tc) < -25:
            raise ValueError('Cannot calculate P Model predictions for values below -25°C. See calc_density_h2o.')

        # Guard against negative VPD issues
        if np.nanmin(self.vpd) < 0:
            raise ValueError('Negative VPD values will lead to missing data - clip to zero or explicitly set to np.nan')

        # ambient CO2 partial pressure (Pa)
        self.ca = calc_co2_to_ca(self.co2, self.patm)

        # photorespiratory compensation point - Gamma-star (Pa)
        self.gammastar = calc_gammastar(tc, patm, pmodel_params=pmodel_params)

        # Michaelis-Menten coef. (Pa)
        self.kmm = calc_kmm(tc, patm, pmodel_params=pmodel_params)

        # viscosity correction factor relative to standards
        self.ns_star = calc_ns_star(tc, patm, pmodel_params=pmodel_params)  # (unitless)

        # Store parameters
        self.pmodel_params = pmodel_params

    def __repr__(self):

        # DESIGN NOTE: This is deliberately extremely terse. It could contain
        # a bunch of info on the environment but that would be quite spammy
        # on screen. Having a specific summary method that provides that info
        # is more user friendly.

        return f"PModelEnvironment(shape={self.shape})"

    def summarize(self, dp=2):
        """Prints a summary of the input and photosynthetic attributes in a
        instance of a PModelEnvironment including the mean, range and number
        of nan values.

        Args:
            dp: The number of decimal places used in rounding summary stats.

        Returns:
            None
        """

        attrs = ['tc', 'vpd', 'co2', 'patm', 'ca', 'gammastar', 'kmm', 'ns_star']
        summarize_attrs(self, attrs, dp=dp)


class PModel:

    r"""Fits the P Model to a given set of environmental and photosynthetic
    parameters. The calculated attributes of the class are described below. An
    extended description with typical use cases is given in :any:`pmodel_overview`
    but the basic flow of the model is:

    1. Estimate :math:`\ce{CO2}` limitation factors and optimal internal to
       ambient :math:`\ce{CO2}` partial pressure ratios (:math:`\chi`), using
       :class:`~pyrealm.pmodel.CalcOptimalChi`.
    2. Estimate photosynthetic efficiencies including water use efficiency,
       light use efficiency (LUE) and maximum carboxylation rate
       (:math:`V_{cmax}`) using :class:`~pyrealm.pmodel.CalcLUEVcmax`.
    3. Optionally, estimate productivity measures including GPP by supplying
       FAPAR and PPFD using the
       :meth:`~pyrealm.pmodel.PModel.estimate_productivity` method.

    **Corollary prediction details**

    These calculations use two additional functions:

    * the instantaneous temperature response of :math:`V_{cmax}`
      (:math:`fv(t)`), implemented in
      :func:`~pyrealm.pmodel.calc_ftemp_inst_vcmax`, and
    * the instantaneous temperature response of dark respiration :math:`V_{d}`
      (:math:`fr(t)`), implemented in
      :func:`~pyrealm.pmodel.calc_ftemp_inst_rd`.

    The predictions are then:

    * Intrinsic water use efficiency (iWUE, Pa), calculated as :math:`(c_a -
      c_i)/1.6`

    * Maximum carboxylation capacity (mol C m-2) normalised to the standard
      temperature as: :math:`V_{cmax25} = V_{cmax}  / fv(t)`

    * Dark respiration, calculated as:

        .. math::

            R_d = b_0 \frac{fr(t)}{fv(t)} V_{cmax}

        following :cite:`Atkin:2015hk` (:math:`b_0` is set in
        :attr:`~pyrealm.pmodel_params.atkin_rd_to_vcmax`)

    * Stomatal conductance (:math:`g_s`), calculated as:

        .. math::

            g_s = \frac{LUE}{M_C}\frac{1}{c_a - c_i}

        When C4 photosynthesis is being used, the true partial pressure of CO2
        in the substomatal cavities (:math:`c_i`) is used following the
        calculation of :math:`\chi` using
        :attr:`~pyrealm.param_classes.PModelParams.beta_cost_ratio_c4`

    * The maximum rate of Rubsico regeneration at the growth temperature
      (:math:`J_{max}`) per unit irradiance is calculated as:

        .. math::

            J_{max} = \frac{4 \phi_0 I_{abs}}{\sqrt{\left(\frac{1}
            {\left(\frac{V_{cmax}(c_i - 2 \Gamma^*)} {\phi_0 I_{abs}(c_i +
            k_{mm})}\right)}\right)^2 - 1}}

    Parameters:

        env: An instance of :class:`~pyrealm.pmodel.PModelEnvironment`. 
        kphio: (Optional) Apparent quantum yield efficiency (unitless). 
        rootzonestress: (Optional, default=None) An experimental option
            for providing a root zone water stress factor. This is not
            compatible with the soilmstress approach.
        soilmstress: (Optional, default=None) A soil moisture stress factor
            calculated using :func:`~pyrealm.pmodel.calc_soilmstress`.
        c4: (Optional, default=False) Selects the C3 or C4 photosynthetic
            pathway. 
        method_jmaxlim: (Optional, default=`wang17`) Method to use for
            :math:`J_{max}` limitation
        do_ftemp_kphio: (Optional, default=True) Include the temperature-
            dependence of quantum yield efficiency (see
            :func:`~pyrealm.pmodel.calc_ftemp_kphio`).

    Attributes:

        env: The photosynthetic environment to which the model is fitted
            (:class:`~pyrealm.pmodel.PModelEnvironment`)
        pmodel_params: The parameters used in the underlying calculations
            (:class:`~pyrealm.param_classes.PModelParams`)
        optchi: Details of the optimal chi calculation
            (:class:`~pyrealm.pmodel.CalcOptimalChi`)
        iwue: Intrinsic water use efficiency (iWUE, Pa) 
        lue: Light use efficiency (LUE)

    Examples:

        >>> env = PModelEnvironment(tc=20, vpd=1000, co2=400, patm=101325.0)
        >>> mod_c3 = PModel(env)
        >>> # Key variables from pmodel
        >>> np.round(mod_c3.optchi.ci, 5)
        28.14209
        >>> np.round(mod_c3.optchi.chi, 5)
        0.69435
        >>> mod_c3.estimate_productivity(fapar=1, ppfd=300)
        >>> np.round(mod_c3.gpp, 5)
        76.42545
        >>> mod_c4 = PModel(env, c4=True, method_jmaxlim='none')
        >>> # Key variables from PModel
        >>> np.round(mod_c4.optchi.ci, 5)
        18.22528
        >>> np.round(mod_c4.optchi.chi, 5)
        0.44967
        >>> mod_c4.estimate_productivity(fapar=1, ppfd=300)
        >>> np.round(mod_c4.gpp, 5)
        103.25886
    """

    def __init__(self,
                 env: PModelEnvironment,
                 rootzonestress: Optional[Union[float, np.ndarray]] = None,
                 soilmstress: Optional[Union[float, np.ndarray]] = None,
                 kphio: Optional[float] = None,
                 do_ftemp_kphio: bool = True,
                 c4: bool = False,
                 method_jmaxlim: str = "wang17",
                 ):

        # Check possible array inputs against the photosynthetic environment
        self.shape = check_input_shapes(env.gammastar, soilmstress, rootzonestress)

        # Store a reference to the photosynthetic environment and a direct
        # reference to the parameterisation
        self.env = env
        self.pmodel_params = env.pmodel_params

        # ---------------------------------------------
        # Soil moisture and root zone stress handling
        # ---------------------------------------------

        if soilmstress is not None and rootzonestress is not None:
            raise AttributeError("Soilmstress and rootzonestress are alternative "
                                 "approaches to soil moisture effects. Do not use both.")

        if soilmstress is None:
            soilmstress = 1.0
            self.do_soilmstress = False
        else:
            self.do_soilmstress = True

        if rootzonestress is None:
            rootzonestress = 1.0
            self.do_rootzonestress = False
        else:
            self.do_rootzonestress = True

        # kphio defaults:
        self.do_ftemp_kphio = do_ftemp_kphio
        if kphio is None:
            if not self.do_ftemp_kphio:
                self.kphio = 0.049977
            elif self.do_soilmstress:
                self.kphio = 0.087182
            else:
                self.kphio = 0.081785
        else:
            self.kphio = kphio

        # -----------------------------------------------------------------------
        # Temperature dependence of quantum yield efficiency
        # -----------------------------------------------------------------------
        # 'do_ftemp_kphio' is not actually a stress function, but is the temperature-
        # dependency of the quantum yield efficiency after Bernacchi et al., 2003

        if self.do_ftemp_kphio:
            self.ftemp_kphio = calc_ftemp_kphio(env.tc, c4,
                                                pmodel_params=env.pmodel_params)
        else:
            self.ftemp_kphio = 1.0

        # -----------------------------------------------------------------------
        # Optimal ci
        # The heart of the P-model: calculate ci:ca ratio (chi) and additional terms
        # -----------------------------------------------------------------------
        self.c4 = c4

        if self.c4:
            method_optci = "c4"
        else:
            method_optci = "prentice14"

        self.optchi = CalcOptimalChi(env.kmm, env.gammastar, env.ns_star,
                                     env.ca, env.vpd, method=method_optci,
                                     rootzonestress=rootzonestress,
                                     pmodel_params=env.pmodel_params)

        # -----------------------------------------------------------------------
        # Vcmax and light use efficiency
        # -----------------------------------------------------------------------

        self.method_jmaxlim = method_jmaxlim
        lue_vcmax = CalcLUEVcmax(self.optchi, self.kphio,
                                 self.ftemp_kphio, soilmstress,
                                 method=method_jmaxlim,
                                 pmodel_params=env.pmodel_params)

        # -----------------------------------------------------------------------
        # Store the two efficiency predictions
        # -----------------------------------------------------------------------

        # intrinsic water use efficiency (in Pa)
        self.iwue = (env.ca - self.optchi.ci) / 1.6
        self.lue = lue_vcmax.lue
        self.vcmax_unit_iabs = lue_vcmax.vcmax

        # -----------------------------------------------------------------------
        # Define attributes populated by estimate_productivity method
        # -----------------------------------------------------------------------
        self._vcmax = None
        self._vcmax25 = None
        self._rd = None
        self._jmax = None
        self._gpp = None
        self._gs = None

    @property
    def gpp(self) -> Union[float, np.ndarray]:
        """Cannot return GPP if estimate_productivity has not been run, do
           not return None silently"""
        if self._gpp is None:
            raise RuntimeError('GPP not calculated: use estimate_productivity')

        return self._gpp

    @property
    def vcmax(self) -> Union[float, np.ndarray]:
        """Cannot return V_cmax if estimate_productivity has not been run, do
           not return None silently"""
        if self._vcmax is None:
            raise RuntimeError('vcmax not calculated: use estimate_productivity')

        return self._vcmax

    @property
    def vcmax25(self) -> Union[float, np.ndarray]:
        """Cannot return V_cmax25 if estimate_productivity has not been run, do
           not return None silently"""
        if self._vcmax25 is None:
            raise RuntimeError('vcmax25 not calculated: use estimate_productivity')

        return self._vcmax25

    @property
    def rd(self) -> Union[float, np.ndarray]:
        """Cannot return RD if estimate_productivity has not been run, do
           not return None silently"""
        if self._rd is None:
            raise RuntimeError('RD not calculated: use estimate_productivity')

        return self._rd

    @property
    def jmax(self) -> Union[float, np.ndarray]:
        """Cannot return Jmax if estimate_productivity has not been run, do
           not return None silently"""
        if self._jmax is None:
            raise RuntimeError('Jmax not calculated: use estimate_productivity')

        return self._jmax

    @property
    def gs(self) -> Union[float, np.ndarray]:
        """Cannot return gs if estimate_productivity has not been run, do
           not return None silently"""
        if self._gs is None:
            raise RuntimeError('GS not calculated: use estimate_productivity')

        return self._gs

    def estimate_productivity(self,
                              fapar: Union[float, np.ndarray] = 1,
                              ppfd: Union[float, np.ndarray] = 1):
        r""" This function takes the light use efficiency and Vcmax per unit
        absorbed irradiance and populates the PModel instance with estimates
        of the following:

            * gpp: Gross primary productivity
            * rd: Dark respiration
            * vcmax: Maximum rate of carboxylation
            * vcmax25: Maximum rate of carboxylation at standard temperature
            * jmax: Maximum rate of electron transport.
            * gs: Stomatal conductance

        The functions finds the total absorbed irradiance (:math:`I_{abs}`) as
        the product of the photosynthetic photon flux density (`ppfd`) and the
        fraction of absorbed photosynthetically active radiation (`fapar`).

        Note that the units of PPFD determine the units of outputs: if PPFD is
        in :math:`\text{mol} m^{-2} \text{month}^{-1}`, then output values are
        scaled per square metre per month.

        The default values of ``ppfd`` and ``fapar`` provide estimates of the
        variables above per unit absorbed irradiance.

        Args:
            fapar: the fraction of absorbed photosynthetically active radiation
            ppfd: photosynthetic photon flux density

        Returns:
            None
        """

        # Check input shapes against each other and an existing calculated value
        _ = check_input_shapes(ppfd, fapar, self.lue)
        
        # Calculate Iabs 
        iabs = fapar * ppfd
        
        # GPP
        self._gpp = self.lue * iabs

        # V_cmax
        self._vcmax = self.vcmax_unit_iabs * iabs

        # V_cmax25 (vcmax normalized to pmodel_params.k_To)
        ftemp25_inst_vcmax = calc_ftemp_inst_vcmax(self.env.tc,
                                                   pmodel_params=self.pmodel_params)
        self._vcmax25 = self.vcmax / ftemp25_inst_vcmax

        # Dark respiration at growth temperature
        ftemp_inst_rd = calc_ftemp_inst_rd(self.env.tc,
                                           pmodel_params=self.pmodel_params)
        self._rd = (self.pmodel_params.atkin_rd_to_vcmax *
                    (ftemp_inst_rd / ftemp25_inst_vcmax) * self.vcmax)

        # Jmax using again A_J = A_C
        fact_jmaxlim = (self.vcmax * (self.optchi.ci + 2.0 * self.env.gammastar) /
                        (self.kphio * iabs * (self.optchi.ci + self.env.kmm)))
        # Guard against negative values getting into sqrt
        jmaxlim_step1 = (1.0 / fact_jmaxlim) ** 2 - 1.0
        jmax = np.empty_like(fact_jmaxlim)
        mask = jmaxlim_step1 > 0
        # Iabs might be a scalar or an array - if an array, it should be of the
        # same shape as jmaxlim_step1 and can use the same mask
        iabs_mask = iabs if isinstance(iabs, (float, int)) else iabs[mask]
        jmax[mask] = 4.0 * self.kphio * iabs_mask / np.sqrt(jmaxlim_step1[mask])
        jmax[~ mask] = np.nan

        # Revert to scalar if needed and store
        self._jmax = jmax.item() if np.ndim(jmax) == 0 else jmax

        # AJ and AC
        a_j = (self.kphio * iabs * (self.optchi.ci - self.env.gammastar) / 
                (self.optchi.ci + 2 * self.env.gammastar) *  fact_jmaxlim)
        a_c = (self.vcmax * (self.optchi.ci - self.env.gammastar)/ 
                (self.optchi.ci + self.env.kmm))
        assim = np.minimum(a_j, a_c)

        if not np.allclose(assim, self._gpp / self.pmodel_params.k_c_molmass, equal_nan=True):
            warn('Assimilation and GPP are not identical')

        # Stomatal conductance
        self._gs = assim / (self.env.ca - self.optchi.ci)

    def __repr__(self):
        if self.do_soilmstress:
            stress = 'Soil moisture'
        elif self.do_rootzonestress:
            stress = 'Root zone'
        else:
            stress = 'None'
        return (f"PModel("
                f"shape={self.shape}, "
                f"kphio={self.kphio}, "
                f"ftemp_kphio={self.do_ftemp_kphio}, "
                f"c4={self.c4}, "
                f"Jmax_method={self.method_jmaxlim}, "
                f"Water stress={stress})")

    def summarize(self, dp=2):
        """Prints a summary of the calculated values in a PModel instance
        including the mean, range and number of nan values. This will always
        show efficiency variables (LUE and IWUE) and productivity estimates are
        shown if {meth}`~pyrealm.pmodel.PModel.calculate_productivity` has been
        run.

        Args:
            dp: The number of decimal places used in rounding summary stats.

        Returns:
            None
        """

        if self._gpp is None:
            attrs = ['lue', 'iwue']
        else:
            attrs = ['lue', 'iwue', 'gpp', 'vcmax',
                     'vcmax25', 'rd', 'gs', 'jmax']

        summarize_attrs(self, attrs, dp=dp)


class CalcOptimalChi:
    r"""Calculate the optimal :math:`\chi` and :math:`\ce{CO2}` limitation
    factors. In more details, the values are:

    - The optimal ratio of leaf internal to ambient :math:`\ce{CO2}` partial
      pressure (:math:`\chi = c_i/c_a`).
    - The :math:`\ce{CO2}` limitation term for light-limited
      assimilation (:math:`m_j`).
    - The :math:`\ce{CO2}` limitation term for Rubisco-limited
      assimilation  (:math:`m_c`).

    The chosen method is automatically used to estimate these values when an
    instance is created.

    Attributes:

        kmm (float): the Michaelis-Menten coefficient (:math:`K`, see
            :func:`calc_kmm`).
        gammastar (float): the photorespiratory :math:`\ce{CO2}` compensation point
            (:math:`\Gamma^{*}`, see :func:`calc_gammastar`).
        ns_star (float): the viscosity correction factor (:math:`\eta^{*}`,
            see :func:`calc_viscosity_H20`)
        ca (float): the ambient partial pressure of :math:`\ce{CO2}` (:math:`c_a`,
            see :func:`calc_co2_to_ca`)
        vpd (float): the vapor pressure deficit (:math:`D`)
        method (str): one of ``c4`` or ``prentice14``
        chi (float): the ratio of leaf internal to ambient :math:`\ce{CO2}`
            partial pressure (:math:`\chi`).
        mj (float): :math:`\ce{CO2}` limitation factor for light-limited
            assimilation (:math:`m_j`).
        mc (float): :math:`\ce{CO2}` limitation factor for RuBisCO-limited
            assimilation (:math:`m_c`).
        mjoc (float):  :math:`m_j/m_c` ratio

    Returns:

        An instance of :class:`CalcOptimalChi` where the :attr:`chi`,
        :attr:`mj`, :attr:`mc` and :attr:`mjoc` have been populated
        using the chosen method.

    Examples:

        >>> vals = CalcOptimalChi(kmm = 46.09928, gammastar = 3.33925,
        ...                       ns_star = 1.12536, ca = 40.53, vpd = 1000)
        >>> round(vals.chi, 5)
        0.69435
        >>> round(vals.mc, 5)
        0.33408
        >>> round(vals.mj, 5)
        0.7123
        >>> round(vals.mjoc, 5)
        2.13211
        >>> # The c4 method estimates chi but sets the others at 1.
        >>> c4_vals = CalcOptimalChi(kmm = 46.09928, gammastar = 3.33925,
        ...                          ns_star = 1.12536, ca = 40.53, vpd = 1000,
        ...                          method='c4')
        >>> round(c4_vals.chi, 5)
        0.44967
    """

    # TODO - move chi calc into __init__? Shared between the two methods

    def __init__(self,
                 kmm: Union[float, np.ndarray],
                 gammastar: Union[float, np.ndarray],
                 ns_star: Union[float, np.ndarray],
                 ca: Union[float, np.ndarray],
                 vpd: Union[float, np.ndarray],
                 rootzonestress: Union[float, np.ndarray] = 1.0,
                 method: str = 'prentice14',
                 pmodel_params: PModelParams = PModelParams()
                 ):

        # Check inputs are broadcastable
        self.shape = check_input_shapes(kmm, gammastar, ns_star,
                                        ca, vpd, rootzonestress)

        # set attribute defaults
        self.chi = None
        self.ci = None
        self.mc = None
        self.mj = None
        self.mjoc = None

        # Identify and run the selected method
        self.pmodel_params = pmodel_params
        self.method = method
        all_methods = {'prentice14': self.prentice14, 'c4': self.c4}

        if self.method in all_methods:
            this_method = all_methods[self.method]
            this_method(kmm=kmm, gammastar=gammastar, ca=ca,
                        vpd=vpd, ns_star=ns_star,
                        rootzonestress=rootzonestress)
        else:
            raise ValueError(f"CalcOptimalChi: method argument '{method}' invalid.")

        # Calculate internal CO2 partial pressure
        self.ci = self.chi * ca

    def __repr__(self):

        return f"CalcOptimalChi(chi={self.chi}, ci={self.ci}, mc={self.mc}, mj={self.mj})"

    def c4(self, **kwargs):
        r"""Optimal :math:`\chi` is calculated following Equation 8 in
        :cite:`Prentice:2014bc` (see :meth:`~pyrealm.pmodel.CalcOptimalChi.prentice14`),
        but using a C4 specific estimate of the unit cost ratio :math:`\beta`,
        specified in :meth:`~pyrealm.param_classes.PModelParams.beta_cost_ratio_c4`.

        This method then simply sets :math:`m_j = m_c = m_{joc} = 1.0` to capture the
        boosted :math:`\ce{CO2}` concentrations at the chloropolast in C4
        photosynthesis.
        """

        # leaf-internal-to-ambient CO2 partial pressure (ci/ca) ratio
        xi = np.sqrt((self.pmodel_params.beta_cost_ratio_c4 *
                      kwargs['rootzonestress'] *
                      (kwargs['kmm'] + kwargs['gammastar']))
                     / (1.6 * kwargs['ns_star']))

        self.chi = (kwargs['gammastar'] / kwargs['ca'] +
                    (1.0 - kwargs['gammastar'] / kwargs['ca']) * xi /
                    (xi + np.sqrt(kwargs['vpd'])))

        # These values need to retain any
        # dimensions of the original inputs - if ftemp_kphio is set to 1.0
        # (i.e. no temperature correction) then the dimensions of tc are lost
        # and the input to soilmstress might be scalar, so enforce the shape.
        # Note that rpmodel_1.0.6 collapses array inputs at this point.

        if self.shape == 1:
            self.mc = 1.0
            self.mj = 1.0
            self.mjoc = 1.0
        else:
            self.mc = np.ones(self.shape)
            self.mj = np.ones(self.shape)
            self.mjoc = np.ones(self.shape)

    def prentice14(self, **kwargs):
        r"""This method calculates key variables as follows:

        Optimal :math:`\chi` is calculated following Equation 8 in
        :cite:`Prentice:2014bc`:

          .. math:: :nowrap:

            \[
                \begin{align*}
                    \chi &= \Gamma^{*} / c_a + (1- \Gamma^{*} / c_a)
                        \xi / (\xi + \sqrt D ), \text{where}\\
                    \xi &= \sqrt{(\beta (K+ \Gamma^{*}) / (1.6 \eta^{*}))}
                \end{align*}
            \]

        The :math:`\ce{CO2}` limitation term of light use efficiency
        (:math:`m_j`) is calculated following Equation 3 in :cite:`Wang:2017go`:

        .. math::

            m_j = \frac{c_a - \Gamma^{*}}
                       {c_a + 2 \Gamma^{*} + 3 \Gamma^{*}
                       \sqrt{\frac{1.6 D \eta^{*}}{\beta(K + \Gamma^{*})}}}

        Finally,  :math:`m_c` is calculated, following Equation 7 in
        :cite:`Stocker:2020dh`, as:

        .. math::

            m_c = \frac{c_i - \Gamma^{*}}{c_i + K},

        where :math:`K` is the Michaelis Menten coefficient of Rubisco-limited
        assimilation.
        """

        # leaf-internal-to-ambient CO2 partial pressure (ci/ca) ratio
        xi = np.sqrt((self.pmodel_params.beta_cost_ratio_c3 *
                      kwargs['rootzonestress'] *
                      (kwargs['kmm'] + kwargs['gammastar']))
                     / (1.6 * kwargs['ns_star']))

        self.chi = (kwargs['gammastar'] / kwargs['ca'] +
                    (1.0 - kwargs['gammastar'] / kwargs['ca']) * xi /
                    (xi + np.sqrt(kwargs['vpd'])))

        # Define variable substitutes:
        vdcg = kwargs['ca'] - kwargs['gammastar']
        vacg = kwargs['ca'] + 2.0 * kwargs['gammastar']
        vbkg = (self.pmodel_params.beta_cost_ratio_c3 *
                kwargs['rootzonestress'] *
                (kwargs['kmm'] + kwargs['gammastar']))

        # Calculate mj
        # NOTE: this differs from rpmodel, which uses length not dim here, so
        # unwrapped matrix inputs. Also, rpmodel includes a check for vpd > 0,
        # but this is guaranteed by clip above (also true in rpmodel).

        vsr = np.sqrt(1.6 * kwargs['ns_star'] * kwargs['vpd'] / vbkg)
        mj = vdcg / (vacg + 3.0 * kwargs['gammastar'] * vsr)

        # Mask values with ns star <= 0 and vbkg <=0
        # TODO - think about masking vs explicit np.nan values in data.
        mj = np.ma.masked_where(np.logical_and(kwargs['ns_star'] <= 0, vbkg <= 0), mj)
        # np.where _always_ returns an array, so catch scalars.
        self.mj = mj.item() if np.ndim(mj) == 0 else mj

        # alternative variables
        gamma = kwargs['gammastar'] / kwargs['ca']
        kappa = kwargs['kmm'] / kwargs['ca']

        # mc and mj:mv
        self.mc = (self.chi - gamma) / (self.chi + kappa)
        self.mjoc = (self.chi + kappa) / (self.chi + 2 * gamma)

    def summarize(self, dp=2):
        """Prints a summary of the variables calculated within an instance
        of CalcOptimalChi including the mean, range and number of nan values.

        Args:
            dp: The number of decimal places used in rounding summary stats.

        Returns:
            None
        """

        attrs = ['chi', 'mc', 'mj', 'mjoc']
        summarize_attrs(self, attrs, dp=dp)


class CalcLUEVcmax:
    r"""Estimate light use efficiency and maximum carboxylation rate
    :math:`V_{cmax}`. The class implements:

    - :math:`J_{max}` limitation of light use efficiency, providing two
      approaches (``wang17`` and ``smith19``),
    - soil moisture stress limitation, and
    - temperature dependence of apparent quantum yield efficiency.

    Light use efficiency (LUE) is calculated from the inputs as:

    .. math::

        \text{LUE} = \phi_0 \cdot \phi_0(T) \cdot  m_j \cdot m_{jlim} \cdot M_C \cdot \beta

    The Rubisco carboxylation capacity (:math:`V_{cmax}`) of the system is then back
    calculated from LUE as:

    .. math::

          V_{cmax} = \frac{\text{LUE}}{m_c M_C}

    Attributes:

        optchi (:class:`CalcOptimalChi`): an instance of :class:`CalcOptimalChi`
            providing the :math:`\ce{CO2}` limitation term of light use efficiency
            (:math:`\m_j`) and the  the :math:`\ce{CO2}` limitation term for
            Rubisco assimilation (:math:`m_c`).
        kphio (float): The apparent quantum yield efficiency (:math:`\phi_0`,
            unitless).
        ftemp_kphio (float): A factor to capture the temperature dependence of
            quantum yield efficiency (:math:`\phi_0(T)`), defaulting to 1.0 for
            no temperature dependence (see :func:`calc_ftemp_kphio`).
        soilmstress (float): A factor to capture the soil moisture stress
            (:math:`\beta`), defaulting to 1.0 for no soil moisture stress
            (see :func:`calc_soilmstress`).
        method (str): method to apply :math:`J_{max}` limitation (default: ``wang17``,
            or ``smith19`` or ``none``)
        mjlim (float): :math:`J_{max}` limitation factor, calculated using the method.
        lue (float): calculated light use efficiency per unit absolute irradiance.
        vcmax (float): calculated maximum carboxylation rate per unit absolute
            irradiance.
        omega (float): component of :math:`J_{max}` calculation (:cite:`Smith:2019dv`).
        omega_star (float):  component of :math:`J_{max}` calculation (:cite:`Smith:2019dv`).

    Other Parameters:

        c_molmass: the molar mass of carbon (:math:`M_C`, `pmodel_params.k_c_molmass`)



    Examples:

        >>> optchi = CalcOptimalChi(kmm = 46.09928, gammastar = 3.33925,
        ...                         ns_star = 1.12536, ca = 40.53, vpd = 1000)
        >>> # Using Wang et al 2017
        >>> out_wang = CalcLUEVcmax(optchi, kphio = 0.081785, ftemp_kphio = 0.656,
        ...                         soilmstress = 1, method='wang17')
        >>> round(out_wang.lue, 5)
        0.25475
        >>> round(out_wang.vcmax, 6)
        0.063488
        >>> # Using Smith et al 2019
        >>> out_smith = CalcLUEVcmax(optchi, kphio = 0.081785, ftemp_kphio = 0.656,
        ...                          soilmstress = 1, method='smith19')
        >>> round(out_smith.lue, 6)
        0.086569
        >>> round(out_smith.vcmax, 6)
        0.021574
        >>> round(out_smith.omega, 5)
        1.10204
        >>> round(out_smith.omega_star, 5)
        1.28251
        >>> # No Jmax limitation
        >>> out_none = CalcLUEVcmax(optchi, kphio = 0.081785, ftemp_kphio = 0.656,
        ...                    soilmstress = 1, method='none')
        >>> round(out_none.lue, 6)
        0.458998
        >>> round(out_none.vcmax, 6)
        0.11439

    """

    # TODO - apparent incorrectness of wang and smith methods with _ca_ variation,
    #        work well with varying temperature but not _ca_ variation (or
    #        e.g. elevation gradient David Sandoval, REALM meeting, Dec 2020)

    def __init__(self, optchi: CalcOptimalChi,
                 kphio: Union[float, np.ndarray],
                 ftemp_kphio: Union[float, np.ndarray] = 1.0,
                 soilmstress: Union[float, np.ndarray] = 1.0,
                 method: str = 'wang17',
                 pmodel_params: PModelParams = PModelParams()
                 ):

        self.shape = check_input_shapes(optchi.mj, optchi.mjoc, kphio,
                                        ftemp_kphio, soilmstress)

        self.optchi = optchi
        self.kphio = kphio
        self.ftemp_kphio = ftemp_kphio
        self.soilmstress = soilmstress
        self.method = method
        self.pmodel_params = pmodel_params
        self.mjlim = None
        self.lue = None
        self.vcmax = None
        self.omega = None
        self.omega_star = None

        all_methods = {'wang17': self.wang17,
                       'smith19': self.smith19,
                       'none': self.none}

        if self.method == 'c4':
            raise ValueError('This class does not implement a fixed method for C4 '
                             'photosynthesis. To replicate rpmodel choose c4=True and '
                             'method="none"')

        if self.method in all_methods:

            # Use the selected method to calculate limitation factors
            this_method = all_methods[self.method]
            this_method()

            # Now calculate the LUE and V_cmax
            # Light use efficiency (gpp per unit absorbed light)
            self.lue = (self.kphio * self.ftemp_kphio * self.optchi.mj * self.mjlim *
                        self.pmodel_params.k_c_molmass * self.soilmstress)

            # Back calculate Vcmax normalised per unit absorbed PPFD (assuming iabs=1)
            self.vcmax = self.lue / (self.optchi.mc * self.pmodel_params.k_c_molmass)

        else:
            raise ValueError(f"CalcLUEVcmax: method argument '{method}' invalid.")

    def __repr__(self):

        return (f"CalcLUEVCmax(lue={self.lue}, vcmax={self.vcmax}, "
                f"mjlim={self.mjlim}, omega={self.omega}, omega_star={self.omega_star})")

    def wang17(self):
        r"""Calculate a :math:`J_{max}` limitation following
        :cite:`Wang:2017go`. The factor is described in Equation 49 of
        :cite:`Wang:2017go` and is the square root term at the end of that
        equation:

        .. math::

            m_{jmax} = \sqrt{1- \left(\frac{c^*}{m_j}\right)^{\frac{2}{3}}}

        Other parameters:

            cstar: A cost parameter for maintaining :math:`J_{max}`
                (:math:`c^*`, `pmodel_params.wang_c`)

        """

        # Calculate mjlim (square root term in Eqn 2 of Wang et al 2017)
        vals = 1 - (self.pmodel_params.wang17_c / self.optchi.mj) ** (2.0 / 3.0)

        # Convert to array if needed and handle negative and nan values
        vals = np.array(vals) if np.ndim(vals) == 0 else vals
        mask = vals >= 0  # Also traps np.nan
        mjlim = np.empty_like(vals)
        mjlim[mask] = np.sqrt(vals[mask])
        mjlim[~ mask] = np.nan

        # revert scalars back to a scalar value
        self.mjlim = mjlim.item() if np.ndim(mjlim) == 0 else mjlim

    def smith19(self):
        r"""Calculate a :math:`J_{max}` limitation following
        :cite:`Smith:2019dv`. The value of :math:`m_{jlim}` is taken as the
        final term of  Equation 18 of :cite:`Smith:2019dv`:

        .. math::
            :nowrap:

            \[
                \begin{align*}
                m_{jlim} &= \frac{\omega^*}{8\theta}, \text{where} \\
                \omega^* &= 1 + \omega - \sqrt{(1 + \omega) ^2 -4\theta\omega}, \text{and}\\
                \omega &= (1 - 2\theta) + \sqrt{(1-\theta)
                    \left(\frac{1}{\frac{4c}{m}(1 - \theta\frac{4c}{m})}-4\theta\right)}
                \end{align*}
            \]

        Other parameters:

            theta: A term to capture the curved relationship between light intensity
                and photosynthetic capacity :math:`J_{max}` (:math:`\theta`, `pmodel_params.smith19_theta`)
            c_cost: A cost parameter for maintaining :math:`J_{max}`
                (:math:`c`, `pmodel_params.smith19_c_cost`)

        """

        # Adopted from Nick Smith's code:
        # Calculate omega, see Smith et al., 2019 Ecology Letters  # Eq. S4
        theta = self.pmodel_params.smith19_theta
        c_cost = self.pmodel_params.smith19_c_cost

        # simplification terms for omega calculation
        cm = 4 * c_cost / self.optchi.mj
        v = 1 / (cm * (1 - self.pmodel_params.smith19_theta * cm)) - 4 * theta

        # account for non-linearities at low m values. This code finds
        # the roots of a quadratic function that is defined purely from
        # the scalar theta, so will always be a scalar. The first root
        # is then used to set a filter for calculating omega.

        cap_p = (((1 / 1.4) - 0.7) ** 2 / (1 - theta)) + 3.4
        aquad = -1
        bquad = cap_p
        cquad = -(cap_p * theta)
        roots = np.polynomial.polynomial.polyroots([aquad, bquad, cquad])

        # factors derived as in Smith et al., 2019
        m_star = (4 * c_cost) / roots[0].real
        omega = np.where(self.optchi.mj < m_star,
                         -(1 - (2 * theta)) - np.sqrt((1 - theta) * v),
                         -(1 - (2 * theta)) + np.sqrt((1 - theta) * v))

        # np.where _always_ returns an array, so catch scalars
        self.omega = omega.item() if np.ndim(omega) == 0 else omega

        self.omega_star = (1.0 + self.omega -  # Eq. 18
                           np.sqrt((1.0 + self.omega) ** 2 -
                                   (4.0 * theta * self.omega)))

        # Effect of Jmax limitation
        self.mjlim = self.omega_star / (8.0 * theta)

    def none(self):
        """No :math:`J_{max}` limitation (:math:`m_{jlim} = 1.0`)
        """

        # Set Jmax limitation to unity - could define as 1.0 in __init__ and
        # pass here, but setting explicitly within the method for clarity.
        self.mjlim = 1.0



import numpy as np
import tabulate
from pyrealm.param_classes import HygroParams
from pyrealm.bounds_checker import bounds_checker
# from pandas.core.series import Series

"""
This module provides utility functions shared by modules or providing
extra functions such as conversions for common forcing variable inputs, 
such as hygrometric and radiation conversions.
"""


def check_input_shapes(*args):
    """This helper function validates inputs to check that they are either
    scalars or arrays and then that any arrays of the same shape. It either
    raises an error or returns the common shape or 1 if all arguments are
    scalar.

    Parameters:

        *args: A set of numpy arrays or scalar values

    Returns:

        The common shape of any array inputs or 1 if all inputs are scalar.

    Examples:

        >>> check_input_shapes(np.array([1,2,3]), 5)
        (3,)
        >>> check_input_shapes(4, 5)
        1
        >>> check_input_shapes(np.array([1,2,3]), np.array([1,2]))
        Traceback (most recent call last):
        ...
        ValueError: Inputs contain arrays of different shapes.
    """

    # Collect the shapes of the inputs
    shapes = set()

    # DESIGN NOTES - currently allow:
    #   - scalars,
    #   - 0 dim ndarrays (also scalars but packaged differently)
    #   - 1 dim ndarrays with only a single value

    for val in args:
        if isinstance(val, np.ndarray):
            # Note that 0-dim ndarrays (which are scalars) pass through as do
            # one dimensional arrays with a single value (also a scalar)
            if not(val.ndim == 0 or val.shape == (1,)):
                shapes.add(val.shape)
        # elif isinstance(val, Series):
        #    # Note that 0-dim ndarrays (which are scalars) pass through
        #    if val.ndim > 0:
        #        shapes.add(val.shape)
        elif val is None or isinstance(val, (float, int, np.generic)):
            pass  # No need to track scalars and optional values pass None
        else:
            raise ValueError(f'Unexpected input to check_input_shapes: {type(val)}')

    # shapes can be an empty set (all scalars) or contain one common shape
    # otherwise raise an error
    if len(shapes) > 1:
        raise ValueError('Inputs contain arrays of different shapes.')

    if len(shapes) == 1:
        return shapes.pop()

    return 1


def summarize_attrs(obj, attrs, dp=2, repr_head=True):
    """
    Helper function to create a simple table of attribute mean, min, max and
    nan count from an object for use in summarize function.

    Args:
        obj: An object with attributes to summarize
        attrs: A list of strings of attribute names
        dp: The number of decimal places used in rounding summary stats.
        repr_head: A boolean indicating whether to show the object representation
          before the summary table.

    Returns:
        None
    """
    
    # Create a list to hold variables and summary stats
    ret = []

    for attr in attrs:
        data = getattr(obj, attr)
        
        # Avoid masked arrays - run into problems with edge cases with all NaN 
        if isinstance(data, np.ma.core.MaskedArray):
            data = data.filled(np.nan)
        
        # Add the variable and stats to the list to be displayed
        ret.append([attr,
                    np.round(np.nanmean(data), dp),
                    np.round(np.nanmin(data), dp),
                    np.round(np.nanmax(data), dp),
                    np.count_nonzero(np.isnan(data))])

    hdrs = ['Attr', 'Mean', 'Min', 'Max', 'NaN']

    if repr_head:
        print(obj)

    print(tabulate.tabulate(ret, headers=hdrs))

# Psychrometric conversions to VPD for vapour pressure, specific humidity and
# relative humidity. Using the bigleaf R package as a checking reference from
# which the doctest values are taken


def calc_vp_sat(ta, hygro_params=HygroParams()):

    """
    Calculates the vapour pressure of saturated air at a given temperature
    in kPa, using the Magnus equation:

    .. math::

        P = a \exp\(\frac{b - T}{T + c}\)

    The parameters :math:`a,b,c` can provided as a tuple, but three
    built-in options can be selected using a string.

    * ``Allen1998``: (610.8, 17.27, 237.3)
    * ``Alduchov1996``: (610.94, 17.625, 243.04)
    * ``Sonntag1990``: (611.2, 17.62, 243.12)

    Args:
        ta: The air temperature
        hygro_params: An object of class ~`pyrealm.param_classes.HygroParams`
            giving the settings to be used in conversions.

    Returns:
        Saturated air vapour pressure in kPa.

    Examples:

        >>> # Saturated vapour pressure at 21°C
        >>> round(calc_vp_sat(21), 6)
        2.480904
        >>> from pyrealm.param_classes import HygroParams
        >>> allen = HygroParams(magnus_option='Allen1998')
        >>> round(calc_vp_sat(21, hygro_params=allen), 6)
        2.487005
        >>> alduchov = HygroParams(magnus_option='Alduchov1996')
        >>> round(calc_vp_sat(21, hygro_params=alduchov), 6)
        2.481888
    """

    # Magnus equation and conversion to kPa
    cf = hygro_params.magnus_coef
    vp_sat = cf[0] * np.exp((cf[1] * ta) / (cf[2] + ta)) / 1000

    return vp_sat


def convert_vp_to_vpd(vp, ta, hygro_params=HygroParams()):
    """Converts vapour pressure to vapour pressure deficit.

    Args:
        vp: The vapour pressure in kPa
        ta: The air temperature in °C
        hygro_params: An object of class ~`pyrealm.param_classes.HygroParams`
            giving the settings to be used in conversions.
    Returns:
        The vapour pressure deficit in kPa

    Examples:
        >>> round(convert_vp_to_vpd(1.9, 21), 7)
        0.5809042
        >>> from pyrealm.param_classes import HygroParams
        >>> allen = HygroParams(magnus_option='Allen1998')
        >>> round(convert_vp_to_vpd(1.9, 21, hygro_params=allen), 7)
        0.5870054
    """
    vp_sat = calc_vp_sat(ta, hygro_params=hygro_params)

    return vp_sat - vp


def convert_rh_to_vpd(rh, ta, hygro_params=HygroParams()):

    """Converts relative humidity to vapour pressure deficit

    Args:
        rh: The relative humidity (proportion in (0,1))
        ta: The air temperature in °C
        hygro_params: An object of class ~`pyrealm.param_classes.HygroParams`
            giving the settings to be used in conversions.
    Returns:
        The vapour pressure deficit in kPa

    Examples:
        >>> round(convert_rh_to_vpd(0.7, 21), 7)
        0.7442712
        >>> from pyrealm.param_classes import HygroParams
        >>> allen = HygroParams(magnus_option='Allen1998')
        >>> round(convert_rh_to_vpd(0.7, 21, hygro_params=allen), 7)
        0.7461016
        >>> import sys; sys.stderr = sys.stdout
        >>> round(convert_rh_to_vpd(70, 21), 7)
        pyrealm/bounds_checker.py:104: UserWarning: Variable rh (proportion) contains values outside the expected range (0,1). Check units?
        -171.1823864
    """

    rh = bounds_checker(rh, 0, 1, '[]', 'rh', 'proportion')

    vp_sat = calc_vp_sat(ta, hygro_params=hygro_params)

    return vp_sat - (rh * vp_sat)


def convert_sh_to_vp(sh, patm, hygro_params=HygroParams()):
    """Convert specific humidity to vapour pressure

    Args:
        sh: The specific humidity in kg kg-1
        patm: The atmospheric pressure in kPa
        hygro_params: An object of class ~`pyrealm.param_classes.HygroParams`
            giving the settings to be used in conversions.
    Returns:
        The vapour pressure in kPa
    Examples:
        >>> round(convert_sh_to_vp(0.006, 99.024), 7)
        0.9517451
    """

    return sh * patm / ((1.0 - hygro_params.mwr) * sh + hygro_params.mwr)


def convert_sh_to_vpd(sh, ta, patm, hygro_params=HygroParams()):
    """Convert specific humidity to vapour pressure deficit

    Args:
        sh: The specific humidity in kg kg-1
        ta: The air temperature in °C
        patm: The atmospheric pressure in kPa
        hygro_params: An object of class ~`pyrealm.param_classes.HygroParams`
            giving the settings to be used in conversions.

    Returns:
        The vapour pressure deficit in kPa

    Examples:
        >>> round(convert_sh_to_vpd(0.006, 21, 99.024), 6)
        1.529159
        >>> from pyrealm.param_classes import HygroParams
        >>> allen = HygroParams(magnus_option='Allen1998')
        >>> round(convert_sh_to_vpd(0.006, 21, 99.024, hygro_params=allen), 5)
        1.53526
    """

    vp_sat = calc_vp_sat(ta, hygro_params=hygro_params)
    vp = convert_sh_to_vp(sh, patm, hygro_params=hygro_params)

    return vp_sat - vp






import numpy as np
from numbers import Number
from typing import Union
from pyrealm import warnings

# DESIGN NOTES: DO 18/08/21
#
# As originally implemented (still commented below), this module provided a
# subclass of numpy.ma.core.MaskedArray. The intention was that a constrained
# array becomes a thing that carries with it a description of its constraints,
# which sounds appealing but...
#
# 1) MaskedArrays have a performance hit.
# 2) Subclasses are 'contagious': so x * subclass returns an object of class
#    subclass. The attributes of that subclass weren't being set and I suspect
#    that is fixable but it doesn't make sense to do so (unless you were mad
#    enough to implement bounds on _all_ values)
# 3) NaN handling. Numpy has np.nan - and masked arrays do handle _masked_
#    np.nan values, but not unmasked ones:
#
#   x = np.ma.masked_array([1,2,np.nan, 4], mask=[1,0,0,1])
#   x.mean()
#   Out[49]: nan
#   x = np.ma.masked_array([1,2,np.nan, 4], mask=[1,0,1,0])
#   x.mean()
#   Out[51]: 3.0
#
# So the general approach here is now:
#
# - a data constraint function that returns simple np.ndarrays or np.nan
# - uses np.nan for both missing / out of range
# - (maybe) has a public mechanism for setting those constraints, rather than
#   hard coding them. But if the ranges are well documented and sane, then
#   this might be fussy, so parking for now.
#
# This is simpler and cleaner but the codebase then needs to systematically
# use nan-aware functions and possibly using bottleneck for speed. Part of
# the problem here is that Numpy lacks a general solution to missing  data:
#   - https://numpy.org/neps/nep-0026-missing-data-summary
#
# Two issues:
# 1) np.nan is a _float_ value and so this can only be applied to float types
#    and hence some type conversion happens.
# 2) The ability to check for already checked data is lost from the previous
#    implementation. This requires the array object to carry attributes, and
#    that implies subclasses.
#
# NOTES: DO 20/11/2021
#
# Actually imposing constraints and enforcing masking is too heavy handed 
# and causes user issues. So, instead retain some of this as a mechanism for
# helping users sanitise inputs, but provide a simple routine to check whether
# values are sane.

def bounds_checker(values: Union[np.ndarray, Number],
                   lower: Number = -np.infty,
                   upper: Number = np.infty,
                   interval_type: str = '[]',
                   label: str = '',
                   unit: str = ''):
    r"""Check inputs fall within bounds

    This is a simple pass through function that tests whether the values fall within
    the bounds specified and issues a warning when this is not the case

    Parameters:
        values: An np.ndarray object or number
        lower: The value of the lower constraint
        upper: The value of the upper constraint
        interval_type: The interval type of the constraint ('[]', '()', '[)', '(]')
        label: A string giving a descriptive label of the variable for use in warnings.
        unit: A string specifying the expected units.

    Returns:
        The function returns the contents of values.

    Examples:

        >>> vals = np.array([-15, 20, 30, 124], dtype=np.float)
        >>> vals_c = bounds_checker(vals, 0, 100, label='temperature', unit='°C')
    """

    # Implement the interval type
    if interval_type not in ['[]', '()', '[)', '(]']:
        raise RuntimeWarning(f'Unknown interval type: {interval_type}')

    if interval_type[0] == '[':
        lower_func = np.less
    else:
        lower_func = np.less_equal

    if interval_type[1] == ']':
        upper_func = np.greater
    else:
        upper_func = np.greater_equal

    # Do the input values contain out of bound values? These tests are not
    # sensitive to dtype, float or int inputs and return either numpy.bool_
    # or np.ndarray with dtype 'bool', both of which support the sum() method
    out_of_bounds = np.logical_or(lower_func(values, lower),
                                  upper_func(values, upper))

    if out_of_bounds.sum():
        warnings.warn(f'Variable {label} ({unit}) contains values outside '
                      f'the expected range ({lower},{upper}). Check units?')

    return values


def input_mask(inputs: Union[np.ndarray, Number],
               lower: Number = -np.infty,
               upper: Number = np.infty,
               interval_type: str = '[]',
               label: str = ''):
    r"""Mask inputs that do not fall within bounds

    This function constrains the values in inputs, replacing values outside
    the provided interval with np.nan. Because np.nan is a float, when any data
    is out of bounds, the returned values are always float arrays or np.nan.

    Parameters:
        inputs: An np.ndarray object or number
        lower: The value of the lower constraint
        upper: The value of the upper constraint
        interval_type: The interval type of the constraint ('[]', '()', '[)', '(]')
        label: A string giving a descriptive label of the constrained contents
            used in reporting.

    Returns:

        If no data is out of bounds, the original inputs are returned, otherwise
        a float np.ndarray object with out of bounds values replaced with np.nan
        or np.nan for number or zero dimension ndarrays.

    Examples:

        >>> vals = np.array([-15, 20, 30, 124], dtype=np.float)
        >>> np.nansum(vals)
        159.0
        >>> vals_c = input_mask(vals, 0, 100, label='temperature')
        >>> np.nansum(vals_c)
        50.0
        >>> vals_c = input_mask(vals, 0, 124, interval_type='[]', label='temperature')
        >>> np.nansum(vals_c)
        174.0
        >>> vals_c = input_mask(vals, 0, 124, interval_type='[)', label='temperature')
        >>> np.nansum(vals_c)
        50.0
    """


    # Implement the interval type
    if interval_type not in ['[]', '()', '[)', '(]']:
        raise RuntimeWarning(f'Unknown interval type: {interval_type}')

    if interval_type[0] == '[':
        lower_func = np.less
    else:
        lower_func = np.less_equal

    if interval_type[1] == ']':
        upper_func = np.greater
    else:
        upper_func = np.greater_equal

    # Do the input values contain out of bound values? These tests are not
    # sensitive to dtype, float or int inputs and return either numpy.bool_
    # or np.ndarray with dtype 'bool', both of which support the sum() method
    mask = np.logical_or(lower_func(inputs, lower),
                         upper_func(inputs, upper))

    # Check if any masking needs to be done
    if mask.sum():

        if isinstance(mask, np.bool_):
            # If mask is np.bool_ then a scalar or zero dimension ndarray was passed,
            # so return np.nan, implicitly converting the input to a float.

            warnings.warn(f"Scalar value {inputs} set to NaN "
                          f"using {interval_type[0]}{lower}, {upper}{interval_type[1]} "
                          f"bounds on {label}",
                          category=RuntimeWarning)

            return np.nan
        elif isinstance(mask, np.ndarray):
            # If an ndarray, then we need a float version to set np.nan and we
            # copy to avoid modifying the original input
            if not np.issubdtype(inputs.dtype, np.floating):
                # Copies implicitly
                outputs = inputs.astype(np.float)
            else:
                outputs = inputs.copy()

            # Count the existing number of NaN values - impossible to have nan
            # in an integer input but isnan works with any input.
            initial_na_count = np.isnan(inputs).sum()

            # Fill in np.nan where values around outside constraints
            outputs[mask] = np.nan

            final_na_count = np.isnan(outputs).sum()

            # Report
            warnings.warn(f"{final_na_count - initial_na_count} values set to NaN "
                          f"using {interval_type[0]}{lower}, {upper}{interval_type[1]} "
                          f"bounds on {label}",
                          category=RuntimeWarning)

            return outputs
        else:
            raise NotImplemented(f'Cannot set bounds on {type(inputs)}')

    else:

        return inputs


# class InputBoundsCheckerFactory:
#     r"""This is a factory used to generate functions to run input_masks
#     with preset limits and labels. The generated functions issue a warning when
#     the constraints are enforced using the provided label to aid reporting.

#     Parameters:

#         lower: The value of the lower constraint
#         upper: The value of the upper constraint
#         interval_type: The interval type of the constraint ('[]', '()', '[)', '(]')
#         label: A string giving a descriptive label of the constrained contents

#     Examples:

#         >>> temp_constraint = InputBoundsCheckerFactory(0, 100, label='temperature (°C)')
#         >>> temp_constraint
#         InputBoundsCheckerFactory: temperature (°C) constrained to [0, 100]
#         >>> vals = np.array([-15, 20, 30, 124], dtype=np.float)
#         >>> np.nansum(vals)
#         159.0
#         >>> vals_c = temp_constraint(vals)
#         >>> np.nansum(vals_c)
#         50.0
#      """

#     def __init__(self,
#                  lower: Number = -np.infty,
#                  upper: Number = np.infty,
#                  interval_type: str = '[]',
#                  label: str = ''):

#         self.label = label
#         self.lower = lower
#         self.upper = upper
#         self.interval_type = interval_type
#         self.interval_notation = f"{interval_type[0]}{lower}, {upper}{interval_type[1]}"

#     def __call__(self, inputs):

#         return input_bounds_checker(inputs, lower=self.lower, upper=self.upper,
#                                     interval_type=self.interval_type,
#                                     label=self.label)

#     def __repr__(self):

#         return f"InputBoundsCheckerFactory: {self.label} constrained to {self.interval_notation}"


# This code was retired in 0.5.3
#
#
# class ConstrainedArray(np.ma.core.MaskedArray):
#     r"""Array with constrained values
#
#     This is a subclass of np.ndarray that constrains the range of permitted
#     values allowed in the array by applying a mask to the original data.
#     This is a general structure to make it easier to apply a constraint and
#     detect when an input has already been constrained. It is based on code
#     from: https://numpy.org/doc/stable/user/basics.subclassing.html
#
#     Parameters:
#
#         input_array: An np.ndarray object
#         lower: The value of the lower constraint
#         upper: The value of the upper constraint
#         interval_type: The interval type of the constraint ('[]', '()', '[)', '(]')
#         label: A string giving a descriptive label of the constrained contents
#
#     Returns:
#
#         A ConstrainedArray object, inheriting from np.ma.core.MaskedArray
#         containing the original data but with values outside the constraints
#         masked.
#
#     Examples:
#
#         >>> vals = np.array([-15, 20, 30, 124])
#         >>> vals.sum()
#         159
#         >>> vals_c = ConstrainedArray(vals, 0, 100, label='temperature')
#         >>> vals_c.sum()
#         50
#         >>> vals_c = ConstrainedArray(vals, 0, 124, interval_type='[]', label='temperature')
#         >>> vals_c.sum()
#         174
#         >>> vals_c = ConstrainedArray(vals, 0, 124, interval_type='[)', label='temperature')
#         >>> vals_c.sum()
#         50
#     """
#
#     def __new__(cls, input_array: np.ndarray,
#                 lower: Number = -np.infty,
#                 upper: Number = np.infty,
#                 interval_type: str = '[]',
#                 label: str = ''):
#
#         # Input array is an already formed ndarray instance
#         # We first cast to be our class type
#         obj = np.asarray(input_array).view(cls)
#
#         # Count the number of already masked cells if any
#         if np.ma.is_masked(obj):
#             n_masked = obj.mask.sum()
#         else:
#             n_masked = 0
#
#         # Implement the interval type
#         if interval_type not in ['[]', '()', '[)', '(]']:
#             raise RuntimeWarning(f'Unknown interval type: {interval_type}')
#
#         if interval_type[0] == '[':
#             lower_func = np.less
#         else:
#             lower_func = np.less_equal
#
#         if interval_type[1] == ']':
#             upper_func = np.greater
#         else:
#             upper_func = np.greater_equal
#
#         # Mask using the constraints as limits
#         mask = np.logical_or(lower_func(obj, lower), upper_func(obj, upper))
#         obj = np.ma.masked_where(mask, obj)
#
#         # Record some details of the constraint
#         obj.lower = lower
#         obj.upper = upper
#         obj.interval_type = interval_type
#         obj.n_constrained = obj.mask.sum() - n_masked
#         obj.label = label
#         obj.interval_notation = f"{interval_type[0]}{lower}, {upper}{interval_type[1]}"
#
#         if obj.n_constrained:
#             warnings.warn(f'{obj.n_constrained} {obj.label} values outside '
#                           f'{obj.interval_notation} masked.',
#                           category=RuntimeWarning)
#         return obj
#
#
# class ConstraintFactory:
#     r"""This is a function factory used to create functions to generate
#     ConstrainedArrays with preset limits and labels. The generated functions
#     issue a warning when the constraints are enforced using the provided label
#     to aid reporting.
#
#     The resulting function also acts a validator. If the input is already
#     a ConstrainedArray, then the function checks that the label and interval
#     match the factory.
#
#     Parameters:
#
#         lower: The value of the lower constraint
#         upper: The value of the upper constraint
#         interval_type: The interval type of the constraint ('[]', '()', '[)', '(]')
#         label: A string giving a descriptive label of the constrained contents
#
#     Examples:
#
#         >>> temp_constraint = ConstraintFactory(0, 100, label='temperature (°C)')
#         >>> temp_constraint
#         ConstraintFactory: temperature (°C) constrained to [0, 100]
#         >>> vals = np.array([-15, 20, 30, 124])
#         >>> vals.sum()
#         159
#         >>> vals_c = temp_constraint(vals)
#         >>> vals_c.sum()
#         50
#         >>> diff = ConstrainedArray(vals, -50, 120, label='foo')
#         >>> diff = temp_constraint(diff)
#         Traceback (most recent call last):
#         ...
#         RuntimeError: Existing input constraints ([-50, 120], foo) do not match checking constraints ([0, 100], temperature (°C))
#         """
#
#     def __init__(self,
#                  lower: Number = -np.infty,
#                  upper: Number = np.infty,
#                  interval_type: str = '[]',
#                  label: str = ''):
#
#         self.label = label
#         self.lower = lower
#         self.upper = upper
#         self.interval_type = interval_type
#         self.interval_notation = f"{interval_type[0]}{lower}, {upper}{interval_type[1]}"
#
#     def __call__(self, arr: np.ndarray):
#
#         if isinstance(arr, ConstrainedArray):
#             if not (arr.lower == self.lower and arr.upper == self.upper and
#                     arr.interval_type == self.interval_type and arr.label == self.label):
#                 raise RuntimeError(f'Existing input constraints ({arr.interval_notation}, {arr.label}) '
#                                    f'do not match checking constraints ({self.interval_notation}, {self.label})')
#             else:
#                 return arr
#
#         arr = ConstrainedArray(arr, lower=self.lower, upper=self.upper,
#                                interval_type=self.interval_type, label=self.label)
#
#         return arr
#
#     def __repr__(self):
#
#         return f"ConstraintFactory: {self.label} constrained to {self.interval_notation}"
#
#



import os
import pytest
import numpy as np
from contextlib import contextmanager
import csv
from pyrealm import tmodel
from pyrealm import param_classes

# ------------------------------------------
# Null context manager to include exception testing in test paramaterisation
# ------------------------------------------

RPMODEL_C4_BUG = True

@contextmanager
def does_not_raise():
    yield


# ------------------------------------------
# Fixtures: inputs and expected values
# ------------------------------------------

@pytest.fixture(scope='module')
def rvalues():
    """Fixture to load test inputs from file.
    """

    test_dir = os.path.dirname(os.path.abspath(__file__))
    with open(os.path.join(test_dir, 'rtmodel_output.csv')) as infile:
        rdr = csv.DictReader(infile, quoting=csv.QUOTE_NONNUMERIC)
        values = [v for v in rdr]

    return values

@pytest.fixture(scope='module')
def pyvalues():
    """Run the same time series in python
    """

    # Do the same simulation as the R implementation - a single stem of
    # diameter 0.1 exposed to 100 years of GPP of 7.
    traits = param_classes.TModelTraits()
    py_vars = ('delta_d', 'diameter', 'height', 'crown_area', 'mass_fol',
               'mass_stm', 'mass_swd', 'gpp_actual', 'resp_swd', 'resp_frt',
               'delta_mass_stm', 'delta_mass_frt')
    values = tmodel.grow_ttree(np.array([[7] * 100]), np.array([0.1]), 1,
                               traits=traits, outvars=py_vars)

    return traits, values

# ------------------------------------------
# Tests
# ------------------------------------------


def test_ttree(rvalues):
    """ Runs a test of the tmodel.TTree against output from an R implementation.
    The values in the test come from simulating a 100 year run starting from
    a stem diameter of 0.1 and with an annual GPP value of 7. Each row in the
    file is the successive growth, but this test just runs each value in
    sequence.

    Args:
        values: Expected outputs from R implementation

    """

    for row in rvalues:

        # create a tree with the initial diameter given in the row
        tree = tmodel.TTree()
        tree.set_diameter(row['D'])
        tree.calculate_growth(7)

        # The R tmodel implementation rescales reported delta_d as
        # a radial increase in millimetres.
        assert np.allclose(tree.delta_d * 500, row['dD'])
        assert np.allclose(tree.diameter, row['D'])
        assert np.allclose(tree.height, row['H'])
        assert np.allclose(tree.crown_area, row['Ac'])
        assert np.allclose(tree.mass_fol, row['Wf'])
        assert np.allclose(tree.mass_stm, row['Ws'])
        assert np.allclose(tree.mass_swd, row['Wss'])
        # The R tmodel implementation slices off foliar respiration costs from
        # GPP before doing anything - the pyrealm.tmodel implementation keeps
        # this cost within the tree calculation
        assert np.allclose(tree.gpp_actual * (1 - tree.traits.resp_f), row['GPP'])
        assert np.allclose(tree.resp_swd, row['Rm1'])
        assert np.allclose(tree.resp_frt, row['Rm2'])
        assert np.allclose(tree.delta_mass_stm, row['dWs'])
        assert np.allclose(tree.delta_mass_frt, row['dWfr'])



@pytest.mark.parametrize(
    'varname',
    [(0, ('dD', 'delta_d')),
     (1, ('D', 'diameter')),
     (2, ('H', 'height')),
     (3, ('Ac', 'crown_area')),
     (4, ('Wf', 'mass_fol')),
     (5, ('Ws', 'mass_stm')),
     (6, ('Wss', 'mass_swd')),
     (7, ('GPP', 'gpp_actual')),
     (8, ('Rm1', 'resp_swd')),
     (9, ('Rm2', 'resp_frt')),
     (10, ('dWs', 'delta_mass_stm')),
     (11, ('dWfr', 'delta_mass_frt'))]
)
def test_grow_ttree(rvalues, pyvalues, varname):
    """ Runs a test of the tmodel.grow_ttree against the same R output.
    In this case, the iteration process through time is tested. If the
    previous test is successful then the values being fed forward in the
    iteration _should_ be all fine, but this checks that the time iteration
    is being run correctly in addition to the implementation of the model being
    correct

    Args:
        values: Expected outputs from R implementation

    """

    idx, (rv, pyv) = varname
    traits, pyvalues = pyvalues

    # Get all the R values across timesteps into an array
    r_var = np.array([rw[rv] for rw in rvalues])
    # Get the matching py output
    py_var = pyvalues[..., idx]

    # Some implementation differences
    if pyv == 'delta_d':
        # The R tmodel implementation rescales reported delta_d as
        # a radial increase in millimetres.
        py_var *= 500
    elif pyv == 'gpp_actual':
        # The R tmodel implementation slices off foliar respiration costs from
        # GPP before doing anything - the pyrealm.tmodel implementation keeps
        # this cost within the tree calculation
        py_var *= (1 - traits.resp_f)

    assert np.allclose(r_var, py_var)



from dataclasses import dataclass, asdict
from typing import Tuple, Union
from numbers import Number
import json
import enforce_typing
from dacite import from_dict

# from box import Box

# Design notes: pyrealm is going to have a bunch of 'deep' settings. Things
# that aren't often tweaked by users but should be easy to tweak when needed.
# The aim here is to have a standard interface to those as an object that maps
# a setting name to a value.
#
# Desired features (in no order):
# 1. Ability to use a obj.attr notation rather than obj['attr']. Not really
#    pythonic but a bit cleaner to read.
# 2. Ability to freeze values. Distinctly paranoid.
# 3. Ability to set a default mapping with default values
# 4. Typing to set expected types on default values.
# 5. Simple export/import methods to go to from dict / JSON
# 6. Is a class, to allow __repr__ and other methods.
#
# ... and then there is a tricky one:
#
# 7. Extensibility. This is the hard one and currently would only be needed
#    to support a customisable version of the T Model. If the T Model could
#    have overridden geometry methods, then these settings _have_ to be able
#    to take extra parameters. And having to set a type on those is another
#    thing that users aren't going to buy into? This makes using @dataclass
#    tricky - because extending class attributes on the fly is really not
#    something that comes naturally to a class. A dotted dict replacement,
#    like Box or addict, is a fairly simpl functional swap.


class ParamClass:
    """Base class for model parameter classes

    This base class provides a consistent interface for creating and exporting
    data from model parameter classes. It defines methods to create parameter
    instances from a dictionary of values and from a JSON file of values.
    It also defines methods to export an existing instance of a parameter
    class to a dictionary or a JSON file.

    All model parameter classes use the :class:`~dataclasses.dataclass`
    class as the basic structure. All fields in these classes are defined
    with default values, so instances can be defined using partial dictionaries
    of alternative values.
    """

    def to_dict(self):
        """Returns a dictionary of the attributes and values used in an instance
        of a parameter class object.

        Returns:
            A dictionary
        """
        return asdict(self)

    def to_json(self, filename):
        """Writes the attributes and values used in an instance of a parameter
        class object to a JSON file.

        Args:
            filename: A path to the JSON file to be created.

        Returns:
            None
        """
        with open(filename, 'w') as outfile:
            json.dump(self.to_dict(), outfile, indent=4)

        return None

    @classmethod
    def from_dict(cls, data):
        """Generates a parameter class object using the data provided in a
        dictionary to override default values

        Args:
            data: A dictionary, keyed by parameter class attribute names,
                  providing values to use in a new instance.

        Returns:
            An instance of the parameter class.
        """
        return from_dict(cls, data)

    @classmethod
    def from_json(cls, filename):
        """Generates a parameter class object using the data provided in a
        JSON file.

        Args:
            filename: The path to a JSON formatted file containing a set
                     of values keyed by parameter class attribute names
                     to use in a new instance.

        Returns:
            An instance of the parameter class.
        """
        with open(filename, 'r') as infile:
            json_data = json.load(infile)

        return cls.from_dict(json_data)


# P Model param class

@enforce_typing.enforce_types
@dataclass(frozen=True)
class PModelParams(ParamClass):

    r"""Model parameters for the P Model

    This dataclass provides a large set of underlying parameters used in
    calculating the predictions of the P Model. The traits are shown below
    with mathematical notation, default value and units shown in brackets:

    **True constants**

    * `k_R`: Universal gas constant (:math:`R` , 8.3145, J/mol/K)
    * `k_co`: O2 partial pressure, Standard Atmosphere (:math:`co` , 209476.0, ppm)
    * `k_c_molmass`: Molecular mass of carbon (:math:`c_molmass` , 12.0107, g)
    * `k_Po`: Standard atmosphere (Allen, 1973)   (:math:`P_o` , 101325.0, Pa)
    * `k_To`: Reference temperature (Prentice, unpublished)   (:math:`T_o` , 25.0, °C)
    * `k_L`: Adiabiatic temperature lapse rate (Allen, 1973)   (:math:`L` , 0.0065, K/m)
    * `k_G`: Gravitational acceleration (:math:`G` , 9.80665, m/s^2)
    * `k_Ma`: Molecular weight of dry air (Tsilingiris, 2008)  (:math:`M_a`, 0.028963, kg/mol)
    * `l_CtoK`: Conversion from °C to K   (:math:`CtoK` , 273.15, -)

    **Density of water**, values taken from Table 5 of :cite:`Fisher:1975tm`.

    * `fisher_dial_lambda`: [1788.316, 21.55053, -0.4695911, 3.096363e-3, -7.341182e-6]
    * `fisher_dial_Po`: [5918.499, 58.05267, -1.1253317, 6.613869e-3, -1.4661625e-5]
    * `fisher_dial_Vinf`: [0.6980547, -7.435626e-4, 3.704258e-5, -6.315724e-7,
      9.829576e-9, -1.197269e-10, 1.005461e-12, -5.437898e-15,
      1.69946e-17, -2.295063e-20]

    **Viscosity of water**, values taken from :cite:`Huber:2009fy`

    * `huber_tk_ast`: Reference temperature (:math:`tk_{ast}`, 647.096, Kelvin)
    * `huber_rho_ast`: Reference density (:math:`\rho_{ast}`, 322.0, kg/m^3)
    * `huber_mu_ast`: Reference pressure (:math:`\mu_{ast}` 1.0e-6, Pa s)
    * `huber_H_i`: Values of H_i (Table 2): (:math:`H_i`, [1.67752, 2.20462, 0.6366564, -0.241605])
    * `huber_H_ij`: Values of H_ij (Table 3): (:math:`H_ij`,
      [[0.520094, 0.0850895, -1.08374, -0.289555, 0.0, 0.0],
      [0.222531, 0.999115, 1.88797, 1.26613, 0.0, 0.120573],
      [-0.281378, -0.906851, -0.772479, -0.489837, -0.257040, 0.0],
      [0.161913,  0.257399, 0.0, 0.0, 0.0, 0.0],
      [-0.0325372, 0.0, 0.0, 0.0698452, 0.0, 0.0],
      [0.0, 0.0, 0.0, 0.0, 0.00872102, 0.0],
      [0.0, 0.0, 0.0, -0.00435673, 0.0, -0.000593264]])

    **Temperature scaling of dark respiration**, values taken from :cite:`Heskel:2016fg`

    * `heskel_b`: Linear factor in scaling (:math:`b`, 0.1012)
    * `heskel_c`: Quadratic factor in scaling (:math:`c`, 0.0005)

    **Temperature and entropy of VCMax**, values taken from Table 3 of :cite:`Kattge:2007db`

    * `kattge_knorr_a_ent`: Offset of entropy vs. temperature relationship (:math:`a_{ent}`, 668.39, J/mol/K)
    * `kattge_knorr_b_ent`: Slope of entropy vs. temperature relationship (:math:`b_{ent}`, -1.07, J/mol/K^2)
    * `kattge_knorr_Ha`: Activation energy (:math:`H_a`, 71513, J/mol)
    * `kattge_knorr_Hd`: Deactivation energy (:math:`H_d`, 200000, J/mol)

    **Scaling of Kphio with temperature**, parameters of quadratic functions

    * `kphio_C4`: Scaling of Kphio in C4 plants, unpublished estimates
      from Shirley (Cai, Wenjia <w.cai17@imperial.ac.uk>) ([-0.064,  0.03, -0.000464])
    * `kphio_C3`: Scaling of Kphio in C3 plants, taken from Table 2 of
      :cite:`Bernacchi:2003dc` ([0.352, 0.022, -3.4e-4])

    **Temperature responses of photosynthetic enzymes**, values taken from Table 1
    of :cite:`Bernacchi:2003dc`

    * dhac: 79430  # (J/mol) Activation energy (Kc)
    * dhao: 36380  # (J/mol) Activation energy (Ko)
    * dha: 37830  # (J/mol) Activation energy (gamma*)
    * # k25 parameters are not dependent on atmospheric pressure, value converted
    * # to Pa by T. Davis assuming elevation of 227.076 m.a.s.l. = 98716.403 Pa
    * kc25: 39.97  # Reported as 404.9 µmol mol-1, converted as 0.0004049 x 98716.403 = 39.97 Pa
    * ko25: 27480  # Reported as 278.4 mmol mol-1, converted as 0.2784 x 98716.403 = 27480 Pa
    * # Reported as 42.75 µmol mol-1,  converted using 42.75 p_0 = 4.332 Pa`
    * gs25_0: 4.332  # Pa

    **Soil moisture stress**, parameterisation from :cite:`Stocker:2020dh`

    * `soilmstress_theta0`: 0.0
    * `soilmstress_thetastar`: 0.6
    * `soilmstress_a`: 0.0
    * `soilmstress_b`: 0.685

    **Unit cost ratio**, value taken from :cite:`Stocker:2020dh`

    * `stocker19_beta`: Unit cost ratio. (:math:`\beta`, 146.0)

    **Electron transport capacity maintenance cost**, value taken from :cite:`Wang:2017go`

    *  `wang_c`: unit carbon cost for the maintenance of electron transport capacity (:math:`c`, 0.41, )

    **Calculation of omega**, scaling factor in J max limitation method of :cite:`Smith:2019dv`:

    * `smith19_theta`: (0.85)
    * `smith19_c_cost`: (0.05336251)

    **Dark respiration**, values taken from :cite:`Atkin:2015hk` for C3 herbaceous plants:

    * `atkin_rd_to_vcmax`:  Ratio of Rdark to Vcmax25 (0.015)
    """

    # Constants
    k_R: Number = 8.3145
    k_co: Number = 209476.0
    k_c_molmass: Number = 12.0107
    k_Po: Number = 101325.0
    k_To: Number = 25.0
    k_L: Number = 0.0065
    k_G: Number = 9.80665
    k_Ma: Number = 0.028963
    k_CtoK: Number = 273.15
    # Fisher Dial
    fisher_dial_lambda: Tuple[Number, ...] = (
        1788.316, 21.55053, -0.4695911, 0.003096363, -7.341182e-06)
    fisher_dial_Po: Tuple[Number, ...] = (
        5918.499, 58.05267, -1.1253317, 0.0066123869, -1.4661625e-05)
    fisher_dial_Vinf: Tuple[Number, ...] = (
        0.6980547, -0.0007435626, 3.704258e-05, -6.315724e-07, 9.829576e-09,
        -1.197269e-10, 1.005461e-12, -5.437898e-15, 1.69946e-17, -2.295063e-20)
    # Huber
    huber_tk_ast: Number = 647.096
    huber_rho_ast: Number = 322.0
    huber_mu_ast: Number =1e-06
    huber_H_i: Tuple[Number,...] = (1.67752, 2.20462, 0.6366564, -0.241605)
    huber_H_ij: Tuple[Tuple[Number, ...], ...] = (
        (0.520094, 0.0850895, -1.08374, -0.289555, 0.0, 0.0),
        (0.222531, 0.999115, 1.88797, 1.26613, 0.0, 0.120573),
        (-0.281378, -0.906851, -0.772479, -0.489837, -0.25704, 0.0),
        (0.161913, 0.257399, 0.0, 0.0, 0.0, 0.0),
        (-0.0325372, 0.0, 0.0, 0.0698452, 0.0, 0.0),
        (0.0, 0.0, 0.0, 0.0, 0.00872102, 0.0),
        (0.0, 0.0, 0.0, -0.00435673, 0.0, -0.000593264))
    # Heskel
    heskel_b: Number = 0.1012
    heskel_c: Number = 0.0005
    # KattgeKnorr
    kattge_knorr_a_ent: Number = 668.39
    kattge_knorr_b_ent: Number = -1.07
    kattge_knorr_Ha: Number = 71513
    kattge_knorr_Hd: Number = 200000
    # Kphio:
    # - note that kphio_C4 has been updated to account for an unintended double
    #   8 fold downscaling to account for the fraction of light reaching PS2.
    #   from original values of [-0.008, 0.00375, -0.58e-4]
    kphio_C4: Tuple[Number, ...] = (-0.064, 0.03, -0.000464)
    kphio_C3: Tuple[Number, ...] = (0.352, 0.022, -0.00034)
    # Bernachhi
    bernacchi_dhac: Number = 79430
    bernacchi_dhao: Number = 36380
    bernacchi_dha: Number = 37830
    bernacchi_kc25: Number = 39.97
    bernacchi_ko25: Number = 27480
    bernacchi_gs25_0: Number = 4.332
    # Soilmstress
    soilmstress_theta0: Number = 0.0
    soilmstress_thetastar: Number = 0.6
    soilmstress_a: Number = 0.0
    soilmstress_b: Number = 0.685
    # Unit cost ratio (beta) (Stocker 2020 value and equivalent for C4).
    beta_cost_ratio_c3: Number = 146.0
    beta_cost_ratio_c4: Number = 16.222222
    # Wang17
    wang17_c: Number = 0.41
    # Smith19
    smith19_theta: Number = 0.85
    smith19_c_cost: Number = 0.05336251
    # Atkin
    atkin_rd_to_vcmax: Number = 0.015


# T model param class

@enforce_typing.enforce_types
@dataclass(frozen=True)
class TModelTraits(ParamClass):
    r"""Trait data settings for a TTree instance

    This data class provides the value of the key traits used in the T model.
    The default values are taken from Table 1 of :cite:`Li:2014bc`. Note that
    the foliage maintenance respiration fraction is not named in the T Model
    description, but has been included as a modifiable trait in this
    implementation. The traits are shown below with mathematical notation,
    default value and units shown in brackets:

    * `a_hd`: Initial slope of height–diameter relationship (:math:`a`, 116.0, -)
    * `ca_ratio`: Initial ratio of crown area to stem cross-sectional area (:math:`c`, 390.43, -)
    * `h_max`: Maximum tree height (:math:`H_m`, 25.33, m)
    * `rho_s`: Sapwood density (:math:`\rho_s`, 200.0, kg Cm−3)
    * `lai`: Leaf area index within the crown (:math:`L`, 1.8, -)
    * `sla`: Specific leaf area (:math:`\sigma`, 14.0, m2 kg−1 C)
    * `tau_f` : Foliage turnover time (:math:`\tau_f`, 4.0, years)
    * `tau_r` : Fine-root turnover time (:math:`\tau_r`, 1.04, years)
    * `par_ext`: PAR extinction coefficient (:math:`k`, 0.5, -)
    * `yld`: Yield_factor (:math:`y`, 0.17, -)
    * `zeta`: Ratio of fine-root mass to foliage area (:math:`\zeta`, 0.17, kg C m−2)
    * `resp_r`: Fine-root specific respiration rate (:math:`r_r`, 0.913, year−1)
    * `resp_s`: Sapwood-specific respiration rate (:math:`r_s`, 0.044, year−1)
    * `resp_f`: Foliage maintenance respiration fraction (:math:`r_f`,  0.1, -)
    """

    a_hd: Number = 116.0        # a, Initial slope of height–diameter relationship (-)
    ca_ratio: Number = 390.43   # c, Initial ratio of crown area to stem cross-sectional area (-)
    h_max: Number = 25.33       # H_m, Maximum tree height (m)
    rho_s: Number = 200.0       # rho_s, Sapwood density (kgCm−3)
    lai: Number = 1.8           # L, Leaf area index within the crown (–)
    sla: Number = 14.0          # sigma, Specific leaf area (m2 kg−1C)
    tau_f: Number = 4.0         # tau_f, Foliage turnover time (years)
    tau_r: Number = 1.04        # tau_r, Fine-root turnover time (years)
    par_ext: Number = 0.5       # k, PAR extinction coefficient (–)
    yld: Number = 0.17          # y, Yield_factor (-)
    zeta: Number = 0.17         # zeta, Ratio of fine-root mass to foliage area (kgCm−2)
    resp_r: Number = 0.913      # r_r, Fine-root specific respiration rate (year−1)
    resp_s: Number = 0.044      # r_s, Sapwood-specific respiration rate (year−1)
    resp_f: Number = 0.1        # --- , Foliage maintenance respiration fraction (-)

    # TODO include range + se, or make this another class TraitDistrib
    #      that can yield a Traits instance drawing from that distribution



@enforce_typing.enforce_types
@dataclass(frozen=True)
class HygroParams(ParamClass):
    r"""Settings for utility function

    This data class provides parameters used :mod:`~pyrealm.utilities`, which
    includes hygrometric conversions

    * `mwr`: The ratio molecular weight of water vapour to dry air (:math:`MW_r`, 0.622, -)
    * `magnus_params`: A three tuple of coefficients for the Magnus equation for
      the calculation of saturated vapour pressure.
    * `magnus_option`: Selects one of a set of published coefficients for the Magnus equation.

    """

    magnus_coef: Tuple[Number, ...] = None
    mwr: Number = 0.622
    magnus_option: str = 'Sonntag1990'

    def __post_init__(self):
        """
        Checks the inputs and populates magnus_coef from the presets if no
        magnus_coef is specified.

        Returns:
            Self
        """
        alts = dict(Allen1998=(610.8, 17.27, 237.3),
                    Alduchov1996=(610.94, 17.625, 243.04),
                    Sonntag1990=(611.2, 17.62, 243.12))

        if self.magnus_coef is None:
            if self.magnus_option not in alts:
                raise (RuntimeError(f"magnus_option must be one of {list(alts.keys())}"))
            else:
                object.__setattr__(self, 'magnus_coef', alts[self.magnus_option])
        elif self.magnus_coef is not None and len(self.magnus_coef) != 3:
            raise (TypeError('magnus_coef must be a tuple of 3 numbers'))
        else:
            object.__setattr__(self, 'magnus_option', None)
